package plugins.big.bigsnake.graph;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import plugins.big.bigsnake.snake.GlobalSnakeSD;
import plugins.big.bigsnake.snake.SubsnakeSD;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DNode;


/**
 * Class that handles a graph-like representation of a GlobalSnake.
 * 
 * @version August 3, 2018
 * 
 * @author Aymeric Galan
 */
public class GlobalSnakeGraph {
	
	private GlobalSnakeSD globalSnake_ = null;
	
	private int numEdges_ = 0;
	private int numEdgesKept_ = 0;
	
	/**
	 * Lists of edges and bounding nodes of the graph.
	 */
	private List<Edge> edgeList_ = new ArrayList<Edge>();
	private List<VertexPair> vertexPairList_ = new ArrayList<VertexPair>();
	
	/**
	 * Lists of edges and bounding nodes of the graph, -without- repeated edges.
	 */
	private List<Edge> edgeListKept_ = new ArrayList<Edge>();
	private List<VertexPair> vertexPairListKept_ = new ArrayList<VertexPair>();
	
	/**
	 * 1D lists for all nodes that were removed when checking duplicated edges
	 * (+ extra list that does not contain staring node of each edge)
	 */
	private NodeList removedNodes_ = null;
//	private NodeList removedNodesNoBounds_ = null;

	private int scalingFactor_ = 2;
	private int subdivLevel_ = 1;

	/**
	 * Mapping between the position of nodes in a given subsnake
	 * and the 'corrected' position in the edge that was not removed
	 * (if ever the node belongs to a duplicated edge)
	 */
	private List<int[][]> subsnakeSubdivisionMap_ = null;

	
	// ========================================================================
	// PUBLIC METHODS
	
	/** Constructor */
	public GlobalSnakeGraph(GlobalSnakeSD globalSnake) {
		globalSnake_ = globalSnake;
		scalingFactor_ = globalSnake_.getScalingFactor();
		subdivLevel_ = globalSnake_.getGlobalLevel();
		
		buildGraph();
		removeDuplicatedEdges();
	}
	
	
	@Override
	public String toString() {
		String eol = System.getProperty("line.separator");
		String numText = "Number of edges = " + numEdges_;
		String numText2 = "Number of kept edges = " + numEdgesKept_;
		return numText + eol + vertexPairList_ + eol + numText2 + eol + vertexPairListKept_;
	}
	
	
	public List<VertexPair> getVertexPairList(boolean noDuplicatedEdge) {
		if (noDuplicatedEdge) return vertexPairListKept_;
		else return vertexPairList_;
	}
	
	
	public List<Edge> getEdgeList(boolean noDuplicatedEdge) {
		if (noDuplicatedEdge) return edgeListKept_;
		else return edgeList_;
	}
	
	
	public NodeList getRemovedNodes() {
		return removedNodes_;
	}
	
	
	public int[][] getSubdivisionMappingForSubsnake(int subsnakeId) {
		return subsnakeSubdivisionMap_.get(subsnakeId);
	}
	
	
	/**
	 * Ask if a given node belongs to a duplicated edge or not
	 * (if yes, this node may then be removed / not tracked in a GlobalSnake)
	 */
	public boolean belongToCleanEdge(Snake2DNode node) {
//	public boolean belongToCleanEdge(Snake2DNode node, boolean noBounds) {
//		if (noBounds) {
//			return !(removedNodesNoBounds_.contains(node));
//		}
//		else {
//			return !(removedNodes_.contains(node));
//		}
		return !(removedNodes_.contains(node));
	}
	
	
	/**
	 * For a given node (which can be a control or a subdivision node), and if this node
	 * belongs to a duplicated edge, return the 'mirror' node that belongs
	 * to the edge that was kept in the graph (its mirror counterpart).
	 * 
	 * ATTENTION : this method returns a <code>Subsnake2DNode</code>, 
	 * which inherits from a <code>Snake2DNode</code> by just adding 
	 * the ID of its <code>SusbnakeCurveSD</code> as a private attribute.
	 */
	public Subsnake2DNode getMirrorNode(Snake2DNode inputNode, int subsnakeId) {
		
		/* if it is not in removed nodes, simply return that node */
		boolean notRemovedNode = !(removedNodes_.contains(inputNode));
		if (notRemovedNode) {
			return new Subsnake2DNode(inputNode, subsnakeId);
		}
		
		/* otherwise, one has to find the 'mirror' node
		 * in the corresponding kept (non-removed) edge
		 */
		
		// 1 - find the edge (that should have been removed in the final graph)
		// that contains the input node, as well as the corresponding vertex pair
		VertexPair pair = null;
		Edge removedEdge = null;
		for (int i = 0 ; i < numEdges_ ; i++) {
			if (edgeList_.get(i).contains(inputNode)) {
				pair = vertexPairList_.get(i);
				removedEdge = edgeList_.get(i);
				break;
			}
		}
		
		// 2 - get the other edge that has the same vertex pair
		Edge keptEdge = edgeListKept_.get(vertexPairListKept_.indexOf(pair));
		
		// 3 - find the position of the input node in the (removed) edge
		int nodeIndexInRemovedEdge = removedEdge.indexOf(inputNode);
		
		// 4 - finally get the 'mirror' node located in the kept edge
		Snake2DNode mirrorNodeTmp = keptEdge.get(nodeIndexInRemovedEdge);
		
		// 5 - assign the ID of the subsnake to the node (return it as a SUBsnake2DNode)
		int id = keptEdge.getParentSnakeId();
		Subsnake2DNode mirrorNode = new Subsnake2DNode(mirrorNodeTmp, id);
		
		return mirrorNode;
	}
	
	
	/**
	 * TODO : add documentation
	 */
	public void buildSubsnakeSubdivisionMapping() {
		
		subsnakeSubdivisionMap_ = new ArrayList<int[][]>();
		
		for (int id = 0 ; id < globalSnake_.getNumSubsnakes() ; id++) {
			
			SubsnakeSD subsnake = globalSnake_.getSubsnake(id);
			
			List<Snake2DNode> subdivNodesList  = subsnake.getInitSubdivisionPoints();
			int numSubdivNodes = subdivNodesList.size();
			
			int[][] oneMap = new int[numSubdivNodes][2];
			
			for (int j = 0 ; j < numSubdivNodes ; j++) {
				
				Snake2DNode node = subdivNodesList.get(j);
				
				Subsnake2DNode correctNode = getMirrorNode(node, id);
				
				int correctId = correctNode.getSubsnakeId();
				
				List<Snake2DNode> correctSubdivNodesList = 
						globalSnake_.getSubsnake(correctId).getInitSubdivisionPoints();
				
				int indexInCorrectSubsnake = correctSubdivNodesList.indexOf(correctNode);
				
				oneMap[j][0] = correctId;
				oneMap[j][1] = indexInCorrectSubsnake;
			}
			
			subsnakeSubdivisionMap_.add(oneMap);
		}
		
	}
	
	
	// ========================================================================
	// PRIVATE METHODS
	
	
	/**
	 * Retrieve subdivision nodes that form an edge linking to node 
	 * in a graph whose nodes are defined by GlobalSnake control nodes.
	 */
	private void buildGraph() {
		edgeList_ = new ArrayList<Edge>();
		vertexPairList_ = new ArrayList<VertexPair>();
		
		// number of subdivision nodes in one edge
		final int twoPowerL = (int) Math.pow(scalingFactor_, subdivLevel_);
		
		List<Snake2DNode> coefAsList = Arrays.asList(globalSnake_.getNodes());
	
		for (int i = 0 ; i < globalSnake_.getNumSubsnakes() ; i++) {
			
			SubsnakeSD subsnake = globalSnake_.getSubsnake(i);
			
			List<Snake2DNode> controlNodesList = subsnake.getInitNodes();
			List<Snake2DNode> subdivNodesList  = subsnake.getInitSubdivisionPoints();
			
			int numControlNodes = controlNodesList.size();
			int numSubdivNodes  = subdivNodesList.size();
			
			for (int j = 0 ; j < numControlNodes ; j++) {
				// build the edge (list of nodes) between two vertices j -> j+1
				int startIdx = j * twoPowerL;
				int endIdx   = (j+1) * twoPowerL;
				
				// grab the nodes to form the edge, excluding first and last node (which are control nodes)
				Edge edge = new Edge(subdivNodesList.subList(startIdx+1, endIdx), i);
				
				// get indices (wrt to coef_ of the j-th Subsnake) of the nodes that form the vertices
				int startCNIdx = controlNodesList.indexOf(subdivNodesList.get(startIdx));
				int endCNIdx   = controlNodesList.indexOf(subdivNodesList.get(endIdx % numSubdivNodes));
				
				// create the nodes
				Snake2DNode startCN = new Snake2DNode(controlNodesList.get(startCNIdx));
				Snake2DNode endCN   = new Snake2DNode(controlNodesList.get(endCNIdx));
				
				// get indices (wrt to coef_ of the GlobalSnake this time)
				int coefStartIdx = coefAsList.indexOf(startCN);
				int coefEndIdx   = coefAsList.indexOf(endCN);
				
				// order the indices in the pair if needed, and revert the node list
				if (coefStartIdx > coefEndIdx) {
					int coefEndIdxBckp = coefEndIdx;
					coefEndIdx = coefStartIdx;
					coefStartIdx = coefEndIdxBckp;
					edge.reverse();
				}
				
				// each edge is bounded by two vertices, stored in tuple (startIdx, endIdx)
				// where indices correspond to those in coef_
				// (the ordering above is to ease the process of duplicates removal)
				VertexPair pair = new VertexPair(coefStartIdx, coefEndIdx);
				
				edgeList_.add(edge);
				vertexPairList_.add(pair);
			}
		}
		numEdges_ = edgeList_.size();
	}
	
	
	/**
	 * This method founds duplicated edges in the graph,
	 * and builds 'clean' versions (= without duplicates) of graph-defining lists,
	 * as well a concatenated list of every nodes that were removed in the process.
	 */
	private void removeDuplicatedEdges() {
		edgeListKept_ = new ArrayList<Edge>();
		vertexPairListKept_ = new ArrayList<VertexPair>();
		removedNodes_ = new NodeList();
		
		for (int i = 0 ; i < numEdges_ ; i++) {
			
			VertexPair pair = vertexPairList_.get(i);
			Edge edge = edgeList_.get(i);
			
			boolean notAlreadySeen = !(vertexPairListKept_.contains(pair));
			if (notAlreadySeen) {
				vertexPairListKept_.add(pair);
				edgeListKept_.add(edge);
			}
			else {
				// add the whole edge
				removedNodes_.addAll(edge);
			}
		}
		numEdgesKept_ = edgeListKept_.size();
	}
	
}
