package plugins.big.bigsnake.jaccard;

import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import icy.main.Icy;
import icy.sequence.Sequence;
import plugins.big.bigsnake.utils.ImageUtils;


/**
 * Very small class that allows the user to specify paths
 * to binary masks, mainly used to compute Jaccard indices.
 * 
 * @version January 18, 2018
 * 
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 *
 */
public class JaccardMaskManager {
	
	/** Number of masks to expect */
	private int numMasks_ = 0;
	
	/**
	 * Memorized directory name where masks are located
	 * so the user has only to specify each one time.
	 */
	private String maskDirectory_ = null;

	/**
	 * Mapping between the mask path and the state 'done' to track which mask 
	 * has already been matched to a position (automatic mode)
	 */
	private Map<String, Boolean> maskPaths_ = null;
	private Map<String, double[][]> maskPixels_ = null;

	private boolean useDefaultDir_ = true;
	
	/**
	 * Hard code your custom mask directory here if you prefer. 
	 * It overrides the <code>maskDirectory_</code> variable.
	 */
	private final static String DEFAULT_MASK_DIR = 
			"/Users/Aymeric/AAAtemp/ideal_emb1_z36_t22/masks/";
	
	/**
	 * Image format, it can can be whatever, 
	 * the only requirement is that masks must be binary 2-bits images 
	 */
	private final static String[] EXTENSIONS = {"jpg", "jpeg", "png", "tif"};
	
	
	// ======================
	// PUBLIC METHODS
	
	/** Constructor. */
	public JaccardMaskManager(int numMasks) {
		numMasks_ = numMasks;
	}

	
	public String getCurrentMaskDirectory() {
		return useDefaultDir_ ? DEFAULT_MASK_DIR : maskDirectory_;
	}
	
	
	public String getMaskDirectory() {
		return maskDirectory_;
	}
	
	
	public String getDefaultDirectory() {
		return DEFAULT_MASK_DIR;
	}
	
	
	public void setMaskDirectory(String dir) {
		maskDirectory_ = dir;
		useDefaultDir_ = false;
	}
	
	public void setMaskDirectoryByUser() {
		final String newMaskDir = requestMaskDirectory();
		maskDirectory_ = newMaskDir;
		useDefaultDir_ = false;
	}
	
	public double[][] getMaskFrom2DPoint(Point2D.Double point) {
		
		if (maskPaths_ == null || maskPixels_ == null) {
			final boolean masksFound = findMasks();
			if (!masksFound) {
				System.err.println("No mask images found in the specified directory");
				return null;
			}
			if (maskPaths_.size() < numMasks_) {
				System.err.println("Not enough mask images in the specified directory");
				return null;
			}
		}
	
		final double x = point.getX();
		final double y = point.getY();
		
		return getMaskPixelsFromPosition((int) x, (int) y);
	}


	public double[][] getMaskFromUser() {
		
		final File chosenImageFile = requestMaskImageFile();
		
		if (chosenImageFile == null) return null;
		
		return getMaskPixelsFromPath(chosenImageFile.getAbsolutePath());
	}
	
	
	public double[][] getMaskPixelsFromPosition(int x, int y) {
		
		double[][] mask = null;
		
		for (Map.Entry<String, Boolean> entry : maskPaths_.entrySet()) {
			
			final String path = entry.getKey();
			final Boolean maskAlreadyDone = entry.getValue();
			
			// get 2D double pixel array
			mask = maskPixels_.get(path);
			
			// take advantage that the mask is binary (0 or 1)
			final boolean pointInsideMask = (mask[x][y] == 1);
			
			if (pointInsideMask) {
				
				if (maskAlreadyDone) {
					System.err.println("Warning : two Subsnakes mapped to the same mask !");
				}
				
				// for debug
//				System.out.println("Mask for position (" 
//						+ x + ", " + y + ") : " + Paths.get(path).getFileName());
				
				// set that mask to the state 'done'
				maskPaths_.put(path, true);
				
				// the mask was found, get out of the loop
				break;
			}
		}
		
		return mask;
	}

	
	public static double[][] getMaskPixelsFromPath(String imageFilePath) {
		return ImageUtils.getPixelsFromPath(imageFilePath, true);
	}
	
	
	// ======================
	// PRIVATE METHODS


	private File requestMaskImageFile() {
		
		JFrame frame = Icy.getMainInterface().getMainFrame();
		
		JFileChooser chooser = null;
		if (useDefaultDir_ && DEFAULT_MASK_DIR != null)
			chooser = new JFileChooser(DEFAULT_MASK_DIR);
		else if (maskDirectory_ != null)
			chooser = new JFileChooser(maskDirectory_);
		else
			chooser = new JFileChooser();
		chooser.setDialogTitle("Select mask image");
		
	    chooser.setFileFilter(getExtensionFilter());
	    
	    int returnVal = chooser.showOpenDialog(frame);
	    if (returnVal == JFileChooser.APPROVE_OPTION) {
	    	File maskImageFile = chooser.getSelectedFile();
	    	maskDirectory_ = maskImageFile.getParent();
	    	return maskImageFile;
	    } else {
	    	return null;
	    }
	}

	
	private String requestMaskDirectory() {
		
		JFrame frame = Icy.getMainInterface().getMainFrame();
		
		JFileChooser chooser = new JFileChooser();
		chooser.setDialogTitle("Select mask directory");
		
		// the user should only be able to select directories (folders)
		chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY); 
		
	    int returnVal = chooser.showOpenDialog(frame);
	    if (returnVal == JFileChooser.APPROVE_OPTION) {
	    	File maskImageDirectory = chooser.getSelectedFile();
	    	return maskImageDirectory.getAbsolutePath();
	    } else {
	    	return null;
	    }
	}
	

	private boolean findMasks() {
		maskPaths_ = new HashMap<String, Boolean>();
		maskPixels_ = new HashMap<String, double[][]>();
		
		final String dir = useDefaultDir_ ? DEFAULT_MASK_DIR : maskDirectory_;
		File[] filesInDirectory = new File(dir).listFiles(getFilenameFilter());
		if (filesInDirectory == null) {
			return false;
		}
		
		for (File file : filesInDirectory) {
		    if (file.isFile()) {
		    	final String path = file.getAbsolutePath();
		    	maskPaths_.put(path, false);
		    	maskPixels_.put(path, getMaskPixelsFromPath(path));
		    }
		}
		return true;
	}
	
	
	private static FileNameExtensionFilter getExtensionFilter() {
		String infoText = "Images ";
		for (String ext : EXTENSIONS) {
			infoText += "*." + ext;
		}
		FileNameExtensionFilter filter = 
				new FileNameExtensionFilter(infoText, EXTENSIONS);
		return filter;
	}
	
	
	private static FilenameFilter getFilenameFilter() {
		FilenameFilter filter = new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				for (String ext : EXTENSIONS) {
					if (name.endsWith(ext)) return true;
				}
				return false;
			}
		};
		return filter;
	}
	
	
	private static void display(BufferedImage img) {
		Sequence sequence = new Sequence();
		sequence.addImage(img);
		sequence.setName("Mask");
		Icy.getMainInterface().addSequence(sequence);
	}
	
}
