package plugins.big.bigsnake.jaccard;

import java.awt.Font;
import java.awt.geom.Point2D;
import java.awt.Dialog.ModalityType;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;

import icy.main.Icy;
import plugins.big.bigsnake.keeper.KeepersListSD;
import plugins.big.bigsnake.keeper.SnakeKeeperSD;
import plugins.big.bigsnake.snake.ESnakeSD;
import plugins.big.bigsnake.snake.SubsnakeSD;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DScale;

public class JaccardSubsnakeStatistics {
	
	/**
	 * Directory where binary mask images are located (Jaccard index).
	 * You can hard code it in the class <code>JaccardMaskManager</code>, 
	 * otherwise it will be memorized by the first time you select a file.
	 */
	private String maskImagesDirectory_ = null;
	
	private KeepersListSD snakeKeepersList_ = null;
	private List<SubsnakeSD> snakeList_ = null;

	private List<Point2D.Double> orderedPositions_ = null;
	
	/** Map that contain Jaccard of its subsnakes and corresponding ID */
	private Map<String, Double> jaccardIndices_ = null;

	/** Statistic fields */
	private int numSubsnakes_ = 0;
	private double average_ = 0;
	private double variance_ = 0;
	private double std_ = 0;
	private String jaccardTextGUI_ = "";
	
	
	public JaccardSubsnakeStatistics(KeepersListSD keepersList, List<Point2D.Double> positionList) {
		snakeKeepersList_ = keepersList;
		orderedPositions_ = positionList;
		numSubsnakes_ = positionList.size();
		jaccardIndices_ = new HashMap<String, Double>();
	}
	
	public JaccardSubsnakeStatistics(List<SubsnakeSD> subsnakeList, List<Point2D.Double> positionList) {
		snakeList_  = subsnakeList;
		orderedPositions_ = positionList;
		numSubsnakes_ = positionList.size();
		jaccardIndices_ = new HashMap<String, Double>();
	}

	
	/** Core computation : get the mask, compute Jaccard indices, then statistics */
	public void run(boolean displayGUI, JaccardIndexMode mode) {
		
		if (snakeKeepersList_ != null) {
			if (numSubsnakes_ > snakeKeepersList_.getNumKeepers()) {
				System.err.println("Missing Subsnakes in the ROI");
				return;
			}
		} else {
			if (numSubsnakes_ > snakeList_.size()) {
				System.err.println("Missing Subsnakes");
				return;
			}
		}
		
		JaccardMaskManager manager = new JaccardMaskManager(numSubsnakes_);
		
		if (maskImagesDirectory_ != null) {
			manager.setMaskDirectory(maskImagesDirectory_);
		}
		
		int idx = 0;
		for (int i = 0 ; i < numSubsnakes_ ; i++) {
			
			SubsnakeSD snake;
			
			if (snakeKeepersList_ != null) {
				
				SnakeKeeperSD keeper = snakeKeepersList_.getKeepers().get(i);
				
				/* activate the current keeper just to help the user
				 * to choose the corresponding mask in the dialog window
				 */
				snakeKeepersList_.activateSnakeKeeper(keeper);
				if (keeper.snake_ instanceof SubsnakeSD) {
					snake = (SubsnakeSD) keeper.snake_;
				} else {
					System.err.println("Please make sure there are only Subsnakes on the sequence");;
					return;
				}
			}
			else {
				snake = (SubsnakeSD) snakeList_.get(i);
			}
		
			double[][] mask;
			
			switch (mode) {
			case AUTOMATIC:
				mask = manager.getMaskFrom2DPoint(orderedPositions_.get(idx));
				break;
			case SEMI_AUTOMATIC:
				if (manager.getMaskDirectory() == null) {
					manager.setMaskDirectoryByUser();
				}
				mask = manager.getMaskFrom2DPoint(orderedPositions_.get(idx));
				break;
			case MANUAL:
				mask = manager.getMaskFromUser();
				break;
			default: // default to manual mode
				mask = manager.getMaskFromUser();
				break;
			}
			
			if (mask == null) {
				System.err.println("Jaccard index computation cancelled.");
				return;
			}
			
			Snake2DScale[] scales = snake.getScales();

			JaccardSnakes jaccard = new JaccardSnakes("", 0, scales, mask);
			final double jaccardIdx = jaccard.run();
			
//				final String id = keeper.getID();
			final String id = String.valueOf(snake.getId());
			
			jaccardIndices_.put(id, jaccardIdx);
			
			if (snakeKeepersList_ != null) {
				/* finally, deactivate the current keeper */
			    snakeKeepersList_.deactivateActiveSnakeKeeper();
			}
			
		    idx++;
		}
		
		computeStatistics();
		
		maskImagesDirectory_ = manager.getCurrentMaskDirectory();
		
		if (displayGUI) {
			displayInIcy();
		}
	}
	
	
	public void setMaskDirectory(String dir) {
		maskImagesDirectory_ = dir;
	}
	
	
	public Map<String, Double> getAllIndices() {
		return jaccardIndices_;
	}
	
	
	public double getAverage() {
		return average_;
	}
	
	
	public double getVariance() {
		return variance_;
	}
	
	
	public double getStandardDeviation() {
		return std_;
	}
	
	
	private static String getLineForPrint(String id, double jaccard) {
		return String.format("Jaccard index for Subsnake %s : %.10f%n", id, jaccard);
	}
	
	
	private static String getLineForGUI(String id, double jaccard) {
		return String.format("J<sub>%s</sub> = %.3f<br>", id, jaccard);
	}
	
	
	private static String getSummaryForGUI(int num, double average, double std) {
		return String.format("<br>Average over %d Subsnakes :<br>"
				+ "<b>J = %.3f \u00B1 %.3f</b>",
				num, average, std);
	}


	private static String getSummaryForPrint(double average, double var, double std) {
		return String.format("Averaged Jaccard index : %.10f \u00B1 %.10f"
				+ " (variance %.6f)%n", 
				average, std, var);
	}
	
	
	private void computeStatistics() {
		
		String jaccardText = "<html><center>";
		
		/* standard average */
		double jaccardIdxAverage = 0;
		for (Map.Entry<String, Double> entry : jaccardIndices_.entrySet()) {
			
			final String id = entry.getKey();
			final Double jaccardIdx = entry.getValue();
			
			final String lineOut = getLineForPrint(id, jaccardIdx);
			System.out.printf(lineOut);
			
			final String lineGUI = getLineForGUI(id, jaccardIdx);
			jaccardText += lineGUI;
			
			jaccardIdxAverage += jaccardIdx;
		}
		// TODO : test weighted average (w.r.t cells area for instance ?)
		jaccardIdxAverage /= (double) numSubsnakes_;
		
		/* standard variance and standard deviation */
		double jaccardIdxVariance = 0;
		for (Double jaccardIdx : jaccardIndices_.values()) {
			jaccardIdxVariance += Math.pow(jaccardIdx - jaccardIdxAverage, 2);
		}
		jaccardIdxVariance /= (double) numSubsnakes_;
		final double jaccardIdxStd = Math.sqrt(jaccardIdxVariance);
		
		average_ = jaccardIdxAverage;
		variance_ = jaccardIdxVariance;
		std_ = jaccardIdxStd;
		
		System.out.printf(getSummaryForPrint(average_, variance_, std_));
		
		jaccardText += getSummaryForGUI(numSubsnakes_, average_, std_);
		jaccardText += "</center></html>";
		jaccardTextGUI_ = jaccardText;
	}


	/**
	 * Display the summary text from Jaccard computation in a dialog box.
	 */
	private void displayInIcy() {
		final JFrame parent = Icy.getMainInterface().getMainFrame();
		final String title = "Jaccard index";
		final JLabel label = new JLabel(jaccardTextGUI_, JLabel.CENTER);
		label.setFont(new Font(label.getFont().getName(), Font.PLAIN, 14));
		
		JOptionPane optionPane = new JOptionPane(label);
	    JDialog dialog = optionPane.createDialog(parent, title);
	    dialog.setModalityType(ModalityType.APPLICATION_MODAL);
	    dialog.setVisible(true);
	}

}
