package plugins.big.bigsnake.nuclei;

import java.awt.geom.Point2D;
import java.awt.geom.Point2D.Double;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import icy.gui.frame.progress.ProgressFrame;
import icy.image.IcyBufferedImage;
import icy.image.IcyBufferedImageUtil;
import icy.main.Icy;
import icy.sequence.Sequence;
import icy.type.DataType;
import plugins.big.bigsnake.roi.ROI2DNucleiDetector;
import plugins.big.bigsnake.tests.nogui.LoG_noGUI;
import plugins.big.bigsnake.utils.FileUtils;
import plugins.big.bigsnake.utils.ImageUtils;
import plugins.big.bigsnakeutils.process.process3D.LoG;

/**
 * LoG_noGUI-based spot detector designed for nuclei detection 
 * in C. elegans embryo LSFM images.
 * 
 * @version January 13, 2018
 * 
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 */
public class NucleiDetector {
	
	private boolean manualMode_ = true;
	private String sequenceName_ = null;
	private Sequence sequence_ = null;
	
	/** Array containing the image data. */
	private double[] imageDataArray_ = null;
	
	/** Array containing a filtered image data smoothed Laplacian filter. */
	private double[] filteredImageDataArray_ = null;
	
	/** Position of nuclei in the nuclei image pixel space. */
	private List<Point2D.Double> nucleiPositions_ = null;
	
	/** Number or nuclei detected. */
	private int numNuclei_ = 0;
	
	/** Width of the original image. */
	private int width_ = 0;
	/** Height of the original image. */
	private int height_ = 0;
	
	/** Filter detectorSigma. */
	private double sigmaLoG_ = 0;
	
	/** 
	 * Threshold for maxima detection
	 * (image should be normalized from NORM_MIN_I to NORM_MAX_I).
	 */
	private double threshold_ = 120;
//	private static final double THRESHOLD = 120;

	/** Variables for normalization of filtered image. */
	private boolean normalizeFilteredImage_ = true;
	private static final double NORM_MIN_I = 0;
	private static final double NORM_MAX_I = 255;
	
	/** Type of connectivity for local max detection. */
	private final ConnectivityType connectType_ = 
			ConnectivityType.FOUR_NEIGHBOR;
	
	/** Initial nuclei position in manual mode. */
	private final ManualPositionType manualPositionType_ = 
			ManualPositionType.TRIANGLE_UP;

	/** ROI. */
	private ROI2DNucleiDetector roiNuclei_ = null;
	
	/** Flag to know is the detection runs in Icy GUI or from the console */
	private boolean withGUI_ = true;
	
	
	// ----------------
	// PUBLIC METHODS
		
	/** Constructor. */
	public NucleiDetector(boolean withGUI) {
		withGUI_ = withGUI;
		if (!withGUI) manualMode_ = false;
	}
	
	
	public void run() {
		if (manualMode_) {
			runManualMode();
			return;
		}

		if (withGUI_) {
			runDetection();
			return;
		}
		else {
			if (sequence_ == null) {
				System.err.println("In non-GUI mode, "
						+ "you must set the sequence manually");
				return;
			}
			runDetectionNoGUI();
		}
	}
	
	
	public boolean isManualMode() {
		return manualMode_;
	}
	
	
	public void setManualMode(boolean manualMode) {
		manualMode_ = manualMode;
	}
	
	
	/**
	 * Setter for the main sequence.
	 */
	public void setSequence(Sequence sequence) {
		sequence_ = sequence;
		updateFields();
	}
	
	
	/** Is initialized ? */
	public boolean isInitialized() {
		if (!manualMode_ && (sigmaLoG_ != 0) && (sequence_ != null)) {
			return true;
		} else if (manualMode_ && (sequence_ != null)) {
			return true;
		}
		return false;
	}
	
	
	/** Nuclei have been detected ? */
	public boolean isDetectionOk() {
		if (nucleiPositions_ != null)
			return (nucleiPositions_.size() > 0);
		return false;
	}
	
	
	/** Getter for nuclei positions. */
	public List<Point2D.Double> getNucleiPositions() {
		return nucleiPositions_;
	}
	
	
	/** Getter for nuclei positions (as an standard array). */
	public Point2D.Double[] getNucleiPositionsAsArray() {
		if (nucleiPositions_ == null)
			return null;
		Point2D.Double[] nucleiPositions = 
				new Point2D.Double[nucleiPositions_.size()];
		nucleiPositions = nucleiPositions_.toArray(nucleiPositions);
		
//		nucleiPositions[0]=new Point2D.Double(106.0, 67.0);
//		nucleiPositions[1]=new Point2D.Double(96.0, 110.0);
//		nucleiPositions[2]=new Point2D.Double(187.0, 97.0);
//		nucleiPositions[3]=new Point2D.Double(141.0, 113.0);
//		nucleiPositions[4]=new Point2D.Double(157.0, 60.0);
		
//		nucleiPositions[0]=new Point2D.Double(222.7210066948286, 52.012690186217455);
//		nucleiPositions[1]=new Point2D.Double(248.06052242657557, 94.02293995200849);
//		nucleiPositions[2]=new Point2D.Double(240.7253994515962, 64.68244805209093);
//		nucleiPositions[3]=new Point2D.Double(221.3873479721051, 82.02001144749676);
//		nucleiPositions[4]=new Point2D.Double(201.3824671312522, 68.68342422026151);
//		nucleiPositions[5]=new Point2D.Double(201.3824671312522, 94.68976931337025);
//		nucleiPositions[6]=new Point2D.Double(192.7136854335493, 116.6951382383084);
//		nucleiPositions[7]=new Point2D.Double(210.05124882895512, 134.03270163371423);
//		nucleiPositions[8]=new Point2D.Double(228.0556415857227, 116.02830887694664);
//		nucleiPositions[9]=new Point2D.Double(176.04295139950523, 74.68488847251737);
//		nucleiPositions[10]=new Point2D.Double(174.70929267678173, 104.02538037243491);
		
//		nucleiPositions[0]=new Point2D.Double(56.00000000000001, 32.0);
//		nucleiPositions[1]=new Point2D.Double(25.6, 68.00000000000001);
//		nucleiPositions[2]=new Point2D.Double(77.20000000000002, 50.0);
//		nucleiPositions[3]=new Point2D.Double(28.4, 39.2);
//		nucleiPositions[4]=new Point2D.Double(51.20000000000001, 59.2);
//		nucleiPositions[5]=new Point2D.Double(61.20000000000001, 75.2);
		
//		nucleiPositions[0]=new Point2D.Double(222.7210066948286, 52.012690186217455);
//		nucleiPositions[1]=new Point2D.Double(248.06052242657557, 94.02293995200849);
//		nucleiPositions[2]=new Point2D.Double(240.7253994515962, 64.68244805209093);
//		nucleiPositions[3]=new Point2D.Double(221.3873479721051, 82.02001144749676);
//		nucleiPositions[4]=new Point2D.Double(201.3824671312522, 68.68342422026151);
//		nucleiPositions[5]=new Point2D.Double(201.3824671312522, 94.68976931337025);
//		nucleiPositions[6]=new Point2D.Double(192.7136854335493, 116.6951382383084);
//		nucleiPositions[7]=new Point2D.Double(210.05124882895512, 134.03270163371423);
//		nucleiPositions[8]=new Point2D.Double(228.0556415857227, 116.02830887694664);
//		nucleiPositions[9]=new Point2D.Double(176.04295139950523, 74.68488847251737);
//		nucleiPositions[10]=new Point2D.Double(174.70929267678173, 104.02538037243491);
		
//		nucleiPositions[0]=new Point2D.Double(152.9837538361505, 82.06025109985084);
//		nucleiPositions[1]=new Point2D.Double(156.50062174042986, 107.26447108051933);
//		nucleiPositions[2]=new Point2D.Double(175.8433952139661, 78.54338319557154);
//		nucleiPositions[3]=new Point2D.Double(161.1897789461356, 59.786754372748476);
//		nucleiPositions[4]=new Point2D.Double(177.01568451539256, 49.82229531062373);
//		nucleiPositions[5]=new Point2D.Double(201.63375984534778, 42.788559502065084);
//		nucleiPositions[6]=new Point2D.Double(226.83797982601627, 49.82229531062373);
//		nucleiPositions[7]=new Point2D.Double(240.31930679242035, 67.99277948273355);
//		nucleiPositions[8]=new Point2D.Double(247.35304260097897, 89.6801315591227);
//		nucleiPositions[9]=new Point2D.Double(229.76870307958237, 116.05664084121763);
//		nucleiPositions[10]=new Point2D.Double(202.21990449606105, 73.85422598986575);
//		nucleiPositions[11]=new Point2D.Double(225.66569052458985, 83.23254040127729);
//		nucleiPositions[12]=new Point2D.Double(205.73677240034036, 94.36928876482848);
//		nucleiPositions[13]=new Point2D.Double(196.94460263964206, 117.22893014264407);
//		nucleiPositions[14]=new Point2D.Double(175.25725056325288, 106.09218177909288);
		
//		nucleiPositions[0]=new Point2D.Double(106.0, 67.0);
//		nucleiPositions[1]=new Point2D.Double(96.0, 110.0);
//		nucleiPositions[2]=new Point2D.Double(187.0, 97.0);
//		nucleiPositions[3]=new Point2D.Double(141.0, 113.0);
//		nucleiPositions[4]=new Point2D.Double(157.0, 60.0);

		
		for(int i=0; i<nucleiPositions.length; i++){
			System.out.println(nucleiPositions[i]);
		}
		return nucleiPositions;
	}
	
	
	/** Getter for the name of sequence to segment. */
	public String getSequenceName() {
		return sequenceName_;
	}

	
	/** Getter for number of detected nuclei. */
	public int getNumNuclei() {
		return numNuclei_;
	}
	
	/** Getter for image width. */
	public int getWidth() {
		return width_;
	}
	
	/** Getter for image height. */
	public int getHeight() {
		return height_;
	}
	
	/** Getter for filtered image. */
	public double[] getFilteredImage() {
		return filteredImageDataArray_;
	}
	
	
//	public double[][] getFilteredImage() {
//		return convert1DTo2DImageRepr(filteredImageDataArray_, 
//				width_, height_);
//	}
	
	
	/** Getter for detectorSigma. */
	public double getSigmaLoG() {
		return sigmaLoG_;
	}
	
	
	/** Setter for detectorSigma. */
	public void setSigmaLoG(double sigma) {
		sigmaLoG_ = sigma;
	}
	
	
	/** Getter for detectorThreshold. */
	public double getThreshold() {
		return threshold_;
	}
	
	
	/** Setter for detectorThreshold. */
	public void setThreshold(double threshold) {
		threshold_ = threshold;
	}
	
	
	/**
	 * Returns false if the nucleus is not present in nucleiPositions_.
	 */
	public boolean removeNucleus(Point2D nucleusPosition) {
		boolean removed = nucleiPositions_.remove(nucleusPosition);
		if (removed) numNuclei_--;
		return removed;
	}
	
	
	/**
	 * Returns false if the nucleus is already present in nucleiPositions_.
	 */
	public boolean addNucleus(Point2D.Double nucleusPosition) {
		final int index = nucleiPositions_.indexOf(nucleusPosition);
		if (index == -1) {
			 nucleiPositions_.add(nucleusPosition);
			 numNuclei_++;
			 return true;
		}
		return false;
	}
	
	
	/** Set a nucleus position to a new one. */
	public void setNewNucleusPosition(Point2D.Double oldPosition, 
			Point2D.Double newPosition) {
		final int index = nucleiPositions_.indexOf(oldPosition);
		if (index == -1) {
			System.err.println("Couldn't update the nucleus position.");
			return;
		}
		nucleiPositions_.set(index, newPosition);
	}
	
	
	/** Mainly for debugging purposes. */
	public void printNuclei() {
		System.out.print("UPDATE : ");
		System.out.println(nucleiPositions_);
		System.out.println("--> nuclei detected : " + nucleiPositions_.size());
	}
	
	
	/** Remove ROI from the sequence. */
	public void removeNucleiROI() {
		sequence_.removeROI(roiNuclei_);
	}
	
	
	/** Display an image in a new sequence. */
	public static void displayImage(double[] image, int width, int height, 
			String sequenceName) {
		Sequence sequence = new Sequence();
		sequence.setName(sequenceName);
		IcyBufferedImage tempImage = new IcyBufferedImage(width, height, 1, DataType.DOUBLE);
		double[] temp = tempImage.getDataXYAsDouble(0);
		
		for(int x = 0 ; x < width ; x++) {
			for (int y = 0; y < height ; y++) {
				temp[x+y*width] = image[x+y*width];
			}
		}
		tempImage.dataChanged();
		sequence.addImage(tempImage);
		
		Icy.getMainInterface().addSequence(sequence);
	}
	
	// ------------------
	// PRIVATE METHODS
	
	private void updateFields() {
		sequenceName_ = sequence_.getName();
		imageDataArray_ = ImageUtils.extractImageFromSequence(sequence_);
		width_ = sequence_.getWidth();
		height_ = sequence_.getHeight();
	}
	
	
	/**
	 * Apply LoG_noGUI filter, then detect local maxima 
	 * and store them as nuclei positions.
	 */
	private void runDetection() {
		if (!isInitialized()) {
			System.err.println("Detector not initialized !");
			return;
		}
		
		ProgressFrame pFrame = new ProgressFrame("Filtering nuclei image...");
		applyFilter();
		pFrame.close();
		
		pFrame = new ProgressFrame("Detecting nuclei...");
		detectNuclei();
		pFrame.close();
		
		roiNuclei_ = new ROI2DNucleiDetector(this);
		
		sequence_ = Icy.getMainInterface().getSequences(sequenceName_).get(0);
		sequence_.addROI(roiNuclei_);
		sequence_.roiChanged();
		
		// DEBUG
//		displayImage(filteredImageDataArray_, width_, height_,
//				sequenceName_ + " (LoG_noGUI-filtered)");
	}
	
	
	/**
	 * Apply LoG_noGUI filter, then detect local maxima 
	 * and store them as nuclei positions.
	 */
	private void runDetectionNoGUI() {
		if (!isInitialized()) {
			System.err.println("Detector not initialized !");
			return;
		}
		applyFilter();
		detectNuclei();
	}
	
	/**
	 * Initialization of manual mode, by default : 
	 * put one nucleus in the middle of the image,
	 * which will be change "by hand" by the user.
	 */
	private void runManualMode() {
		if (!isInitialized()) {
			System.err.println("Detector not initialized !");
			return;
		}
		
		switch(manualPositionType_) {
		case TWO_HORIZONTAL:
			nucleiPositions_ = SeedPositioner.horizontalTwo(width_, height_);
			break;
		case TWO_VERTICAL:
			nucleiPositions_ = SeedPositioner.verticalTwo(width_, height_);
			break;
		case TRIANGLE_UP:
			nucleiPositions_ = SeedPositioner.triangleUp(width_, height_);
			break;
		default:
			nucleiPositions_ = SeedPositioner.triangleUp(width_, height_);
			break;
		}
		numNuclei_ = nucleiPositions_.size();
		
		roiNuclei_ = new ROI2DNucleiDetector(this);
		sequence_.addROI(roiNuclei_);
		sequence_.roiChanged();
	}
	
	
	/**
	 * Add "_nuclei" to a file name.
	 */
	private String getNameWithNuclei(String nameToSegment) {
		String[] baseNameAndExt = FileUtils.removeAndGetExtension(nameToSegment);
		String nameWithNuclei = baseNameAndExt[0] + "_nuclei" + baseNameAndExt[1];
		return nameWithNuclei;
	}
	
	
	/**
	 * Apply Laplacian of Gaussian (LoG_noGUI) filter to image with nuclei.
	 */
	private void applyFilter() {
		/* Using LoG3D but actually in 2D */
		double[][] pseudo3DImage = new double[1][height_*width_];
		for (int k = 0 ; k < width_*height_ ; k++) {
			pseudo3DImage[0][k] = imageDataArray_[k];
		}
		
		double[][] filtered3DImage = null;
		if (withGUI_) {
			LoG LoGFilter = new LoG(pseudo3DImage, 
					width_, height_, 1, sigmaLoG_);
			filtered3DImage = LoGFilter.filter();
		} else {
			LoG_noGUI LoGFilter = new LoG_noGUI(pseudo3DImage, 
					width_, height_, 1, sigmaLoG_);
			filtered3DImage = LoGFilter.filter();
		}
	
		filteredImageDataArray_ = new double[width_*height_];
		for (int k = 0 ; k < width_*height_ ; k++) {
			filteredImageDataArray_[k] = filtered3DImage[0][k];
		}
		
		// intensity normalization
		if (normalizeFilteredImage_) {
			// first compute here the min and max pixel intensities
			double maxI = -java.lang.Double.MAX_VALUE;
			double minI = java.lang.Double.MAX_VALUE;
			for (int k = 0 ; k < width_*height_ ; k++) {
				double I = filteredImageDataArray_[k];
			    if (I < minI) minI = I;
			    if (I > maxI) maxI = I;
			}
			
			// then perform the actual normalization
			for (int k = 0 ; k < width_*height_ ; k++) {
				double I = filteredImageDataArray_[k];
				double normI = (I - minI) * (NORM_MAX_I - NORM_MIN_I) /
						(maxI - minI) + NORM_MIN_I;
				filteredImageDataArray_[k] = normI;
			}
		}
	}
	
	
	private void detectNuclei() {
		nucleiPositions_ = detectLocalMaxima(filteredImageDataArray_,
				width_, height_, threshold_, connectType_);
		numNuclei_ = nucleiPositions_.size();
	}
	
	
	/**
	 * Detect local maxima in a image array using 4- or 8- neighbor connectivity,
	 * given an input image, its width and height, a detectorThreshold,
	 * and a type of connectivity (8-neighbor for instance).
	 */
	private static ArrayList<Point2D.Double> detectLocalMaxima(double[] inputImage, 
			int nx, int ny, double threshold, ConnectivityType connectType) {
		
		ArrayList<Point2D.Double> maximaPositions = new ArrayList<Point2D.Double>();
		
		// to avoid pixels out of image bounds to be treated as local maxima, 
		// we initialize neighborhood to that dumb negative value.
		final double outOfBoundsPixelValue = -1;
		
		for (int x = 0 ; x < nx ; x++) {
			for (int y = 0 ; y < ny ; y++) {
				
				double pixelValue = inputImage[x + y*nx];
				double[] neighborPixelValues = null;
				
				switch (connectType) {
				case FOUR_NEIGHBOR:
					neighborPixelValues = new double[4];
					break;
				case EIGHT_NEIGHBOR:
					neighborPixelValues = new double[8];
					break;
				default:
					System.err.println("Unsupported connectivity -> "
							+ "default to 4-neighbor.");
					neighborPixelValues = new double[4];
					break;
				}
				getPixelNeighborhood(inputImage, x, y, nx, ny, 
						outOfBoundsPixelValue, neighborPixelValues);
				
				if (isLocalMaximum(pixelValue, neighborPixelValues) 
						&& (pixelValue > threshold)) {
					maximaPositions.add(new Point2D.Double(x, y));
				}
			}
		}
		
		return maximaPositions;
	}


	private static void getPixelNeighborhood(double[] inputImage, 
			int x, int y, int nx, int ny, double undefPixelValue,
			double[] neighborPixelValues) {
		
		Arrays.fill(neighborPixelValues, undefPixelValue);
		
		// TODO: maybe pass a chunk (3x3 here) of the inputImage 
		// in place of the whole image ?
		
		if (neighborPixelValues.length == 4) {
			get4Neighbor(inputImage, x, y, nx, ny, neighborPixelValues);
		} else if (neighborPixelValues.length == 8) {
			get8Neighbor(inputImage, x, y, nx, ny, neighborPixelValues);
		}
	}
	
	private static void get4Neighbor(double[] inputImage, 
			int x, int y, int nx, int ny, double[] outputNeigborhood) {
		if (isNotFirst(y, ny)) {
			outputNeigborhood[0] = inputImage[x + (y-1)*nx]; // top pixel
		}
		if (isNotFirst(x, nx)) {
			outputNeigborhood[1] = inputImage[(x-1) + y*nx]; // left pixel
		}
		if (isNotLast(x, nx) ) {
			outputNeigborhood[2] = inputImage[(x+1) + y*nx]; // right pixel
		}
		if (isNotLast(y, ny)) {
			outputNeigborhood[3] = inputImage[x + (y+1)*nx]; // bottom pixel
		}
	}
	
	private static void get8Neighbor(double[] inputImage, 
			int x, int y, int nx, int ny, double[] outputNeigborhood) {
		if (isNotFirst(x, nx) && isNotFirst(y, ny)) {
			outputNeigborhood[0] = inputImage[(x-1) + (y-1)*nx]; // top left pixel
		}
		if (isNotFirst(y, ny)) {
			outputNeigborhood[1] = inputImage[x + (y-1)*nx]; // top pixel
		}
		if (isNotLast(x, nx) && isNotFirst(y, ny)) {
			outputNeigborhood[2] = inputImage[(x+1) + (y-1)*nx]; // top right pixel
		}
		if (isNotFirst(x, nx)) {
			outputNeigborhood[3] = inputImage[(x-1) + y*nx]; // left pixel
		}
		if (isNotLast(x, nx)) {
			outputNeigborhood[4] = inputImage[(x+1) + y*nx]; // right pixel
		}
		if (isNotFirst(x, nx) && isNotLast(y, ny)) {
			outputNeigborhood[5] = inputImage[(x-1) + (y+1)*nx]; // bottom left pixel
		}
		if (isNotLast(y, ny)) {
			outputNeigborhood[6] = inputImage[x + (y+1)*nx]; // bottom pixel
		}
		if (isNotLast(x, nx) && isNotLast(y, ny)) {
			outputNeigborhood[7] = inputImage[(x+1) + (y+1)*nx]; // bottom right pixel
		}
	}
	
	
	private static boolean isInside(int x, int y, int nx, int ny) {
		return (isNotFirst(x, nx) && isNotLast(x, nx) && 
				isNotFirst(y, ny) && isNotLast(y, ny));
	}
	
	
	private static boolean isNotFirst(int i, int ni) {
		return (i != 0);
	}
	
	
	private static boolean isNotLast(int i, int ni) {
		return (i != ni-1);
	}
	
	
	private static boolean isLocalMaximum(double pixelValue, 
			double[] neighborPixelValues) {
		for(int k = 0 ; k < neighborPixelValues.length ; k++) {
			if (pixelValue < neighborPixelValues[k]) return false;
		}
		return true;
	}

}
