package plugins.big.bigsnake.roi;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.util.ArrayList;

import javax.swing.JOptionPane;

import icy.canvas.Canvas2D;
import icy.canvas.IcyCanvas;
import icy.canvas.IcyCanvas2D;
import icy.gui.dialog.MessageDialog;
import icy.painter.Anchor2D;
import icy.painter.PathAnchor2D;
import icy.sequence.Sequence;
import icy.type.point.Point5D;
import icy.util.EventUtil;
import plugins.big.bigsnake.nuclei.NucleiDetector;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DScale;
import plugins.kernel.canvas.VtkCanvas;
import plugins.kernel.roi.roi2d.ROI2DShape;

public class ROI2DNucleiDetector extends ROI2DShape {
	
	private NucleiDetector detector_ = null;
	
//	private Point2D.Double[] nucleiPositions_ = null;
	
	private boolean isEditable_ = true;
	
	private boolean hasChanged_ = false;
	
	/** Cursor shown on the screen when dragging nucleus marker. */
	private final Cursor dragNucleusCursor_ = Cursor
			.getPredefinedCursor(Cursor.MOVE_CURSOR);
	/** Cursor shown on the screen when adding nucleus marker. */
	private final Cursor addNucleusCursor_ = Cursor
			.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR);
	
	/** Color of nuclei markers. */
	private static final Color DEFAULT_NORMAL_COLOR = Color.RED;
	private static final Color DEFAULT_SELECTED_COLOR = Color.GREEN;
	
	/** Size of nuclei markers. */
	private static final int MARKER_SIZE = 6;
	
	// --------------------
	// PUBLIC METHODS

	public ROI2DNucleiDetector(NucleiDetector detector) {
		super(new Path2D.Double());
		
		detector_ = detector;
		setName("nuclei");
		
		for (Point2D.Double pos : detector_.getNucleiPositions()) {
			Anchor2D pt = createAnchor(pos);
            addPoint(pt);
		}
		
		// but we only want points to be displayed, not the path
//		setSelected(true);
	}
	
	
	// ------------------
	// PRIVATE METHODS
	
	private ArrayList<Anchor2D> getAnchorsFromShape(Shape shape) {
		final PathIterator pathIt = shape.getPathIterator(null);
		final ArrayList<Anchor2D> result = new ArrayList<Anchor2D>();
		final double crd[] = new double[6];
		while (!pathIt.isDone()) {
			final int segType = pathIt.currentSegment(crd);
			PathAnchor2D pt = null;
			if (segType != PathIterator.SEG_CLOSE) {
				pt = (PathAnchor2D) createAnchor(new Point2D.Double(crd[0],
						crd[1]));
				pt.setType(segType);
				result.add(pt);
			}
			pathIt.next();
		}
		return result;
	}
	
	
	@Override
	protected Anchor2D createAnchor(Point2D pos) {
		return new NucleiMarkerPainter(pos, DEFAULT_NORMAL_COLOR,
				DEFAULT_SELECTED_COLOR, MARKER_SIZE);
	}
	
	
	@Override
	protected ROI2DShapePainter createPainter() {
		return new ShapePainter();
	}
	
	// ============================================================================
	// INNER CLASSES

	/**
	 * Class that handles the interaction and the display of the control points
	 * of the snake.
	 * 
	 * @version November 27, 2017
	 * 
	 * @author Aymeric Galan (aymeric.galan@gmail.com)
	 */
	private class NucleiMarkerPainter extends PathAnchor2D {

		// ============================================================================
		// PUBLIC METHODS

		/** Constructor. */
		public NucleiMarkerPainter(Point2D pos, Color defaultSelectedColor,
				Color overColor, int markerSize) {
			super(pos.getX(), pos.getY(), markerSize,
					defaultSelectedColor, overColor);
		}

		// ----------------------------------------------------------------------------

		/** Draws the cross of the control point on the canvas. */
		@Override
		public void paint(Graphics2D g, Sequence sequence, IcyCanvas canvas) {
	
			if (canvas instanceof IcyCanvas2D) {
				final double adjRayX = canvas.canvasToImageLogDeltaX(ray);
				final double adjRayY = canvas.canvasToImageLogDeltaY(ray);
				double stroke = 1.5 * getAdjustedStroke(canvas);
				g.setStroke(new BasicStroke((float) stroke));

				if (selected) {
					g.setColor(selectedColor);
				} else {
					g.setColor(color);
				}
				g.draw(new Line2D.Double(position.x - adjRayX, position.y,
						position.x + adjRayX, position.y));
				g.draw(new Line2D.Double(position.x, position.y - adjRayY,
						position.x, position.y + adjRayY));
			} else {
				System.err.println("Unrecognized IcyCanvas.");
			}
		}
	}
	
	
	// ----------------------------------------------------------------------------

	/**
	 * Class that (un)draws the ROI forming the inside of the path (we only need points).
	 * 
	 * @version November 27, 2017
	 * 
	 * @author Aymeric Galan (aymeric.galan@gmail.com)
	 */
	private class ShapePainter extends ROI2DShapePainter {
		
		/** Anchor2D point that is currently being selected/dragged. */
		private Anchor2D selectedPt_ = null;
		
		/** Radius (in pixels) to consider the mouse is over a marker. */
		private static final double EPSILON = (double) MARKER_SIZE;

		// ============================================================================
		// PUBLIC METHODS

		/**
		 * Invoked when a key is pressed. This method is used to block the key
		 * events from Icy.
		 */
		@Override
		public void keyPressed(KeyEvent e, Point5D.Double imagePoint,
				IcyCanvas canvas) {
			e.consume();
		}
		
		/**
		 * Invoked when the mouse cursor has been moved onto the Icy canvas but
		 * no buttons have been pushed.
		 */
		@Override
		public void mouseMove(MouseEvent e, Point5D.Double imagePoint,
				IcyCanvas canvas) {
			
			if (isEditable_) {
				
				if (!isActiveFor(canvas)) return;
				
				ROI2DNucleiDetector.this.beginUpdate();
				try {
					for (Anchor2D pt : controlPoints) {
						if (pt.isOver(canvas, imagePoint.getX(), imagePoint.getY())) {
							((Canvas2D) canvas).getViewComponent().setCursor(
									addNucleusCursor_);
						}
					}
					
					for (Anchor2D pt : controlPoints) {
						pt.mouseMove(e, imagePoint, canvas);
					}
				} finally {
					ROI2DNucleiDetector.this.endUpdate();
				}
				
			}
		}
		
		
		/** Invoked when a mouse button has been pressed on the Icy canvas. */
		@Override
		public void mousePressed(MouseEvent e, Point5D.Double imagePoint,
				IcyCanvas canvas) {
			
			if (e.isConsumed()) return;
			
			if (!isActiveFor(canvas)) return;
			
			if (isEditable_) {
				ROI2DNucleiDetector.this.beginUpdate();
				try {
					// right click to remove marker
					if (EventUtil.isRightMouseButton(e)) {
						removeMarker(imagePoint, canvas);
					} 
					// left click to add marker
					else if (EventUtil.isLeftMouseButton(e)) {
						addNewMarker(imagePoint, canvas);
					}
					
					for (Anchor2D pt : controlPoints) {
						if (pt.isSelected()) {
							selectedPt_ = new Anchor2D(pt.getX(), pt.getY());
						}
					}
					
					e.consume();
				} finally {
					ROI2DNucleiDetector.this.endUpdate();
				}
			}
		}
		
		
		/** Invoked when a mouse button has been pressed on the Icy canvas. */
		@Override
		public void mouseReleased(MouseEvent e, Point5D.Double imagePoint,
				IcyCanvas canvas) {
			
			if (e.isConsumed()) return;
			
			if (!isActiveFor(canvas)) return;
				
			if (isEditable_) {
				
				ROI2DNucleiDetector.this.beginUpdate();
				try {
					for (Anchor2D pt : controlPoints) {
						if (pt.isSelected()) {
							Point2D.Double oldPos = new Point2D.Double(
									selectedPt_.getX(), 
									selectedPt_.getY());
							Point2D.Double newPos = new Point2D.Double(
									pt.getX(), pt.getY());
							
							// update the detector's list
							detector_.setNewNucleusPosition(oldPos, newPos);
						}
					}
					
					e.consume();
				} finally {
					ROI2DNucleiDetector.this.endUpdate();
				}
			}
		}
		
		/**
		 * Invoked when a mouse button is pressed on the Icy canvas and then
		 * dragged.
		 */
		@Override
		public void mouseDrag(MouseEvent e, Point5D.Double imagePoint,
				IcyCanvas canvas) {
			
			if (e.isConsumed()) return;
			
			if (!isActiveFor(canvas)) return;
			
			if (isEditable_) {
				
				ROI2DNucleiDetector.this.beginUpdate();
				try {
					for (Anchor2D pt : controlPoints) {
						pt.mouseDrag(e, imagePoint, canvas);
					}
					e.consume();	
				} finally {
					ROI2DNucleiDetector.this.endUpdate();
				}	
			}
		}
		
		// ============================================================================
		// PROTECTED METHODS

		/**
		 * Draws on the image canvas only the control points (i.e. nuclei positions)
		 * instead o the whole path and everything inside.
		 */
		@Override
		protected void drawROI(Graphics2D g, Sequence sequence, IcyCanvas canvas) {
			if (canvas instanceof IcyCanvas2D) {
				final Graphics2D g2 = (Graphics2D) g.create();
				for (Anchor2D pt : controlPoints) {
					pt.paint(g2, sequence, canvas);
				}
				g2.dispose();
			} else if (canvas instanceof VtkCanvas) {
				super.drawROI(g, sequence, canvas);
			}
		}

		
		// ============================================================================
		// PRIVATE METHODS
		
		private void addNewMarker(Point5D.Double imagePoint, IcyCanvas canvas) {
			// for practicality, get the 2D version of the 5D imagePoint
			Point2D.Double imagePoint2D = new Point2D.Double(imagePoint.getX(), 
					imagePoint.getY());
			
			/* first find the closest marker to imagePoint (within a radius EPSILON) */
			// starts with first point (index 0)
			double closestDistance = imagePoint2D.distance(
					controlPoints.get(0).getPosition());
			for (int i = 1 ; i < controlPoints.size() ; i++) {
				double dist = imagePoint2D.distance(
						controlPoints.get(i).getPosition());
				if (dist < closestDistance) {
					closestDistance = dist;
				}
			}
			
			// prevent adding a marker too close to another one
			if (closestDistance < EPSILON) return;
			
			ROI2DNucleiDetector.this.beginUpdate();
			try {
				// get the position of the click
				Point2D.Double pos = new Point2D.Double(imagePoint.getX(), 
						imagePoint.getY());
				
				// add the new marker
				final Anchor2D pt = createAnchor(pos);
	            addPoint(pt);
	            pt.setSelected(true);
	            
	            // update the detector's list
				detector_.addNucleus(pos);
				
				hasChanged_ = true;
			} finally {
				ROI2DNucleiDetector.this.endUpdate();
			}
		}

		
		private void removeMarker(Point5D.Double imagePoint, IcyCanvas canvas) {
			// for practicality, get the 2D version of the 5D imagePoint
			Point2D.Double imagePoint2D = new Point2D.Double(imagePoint.getX(), 
					imagePoint.getY());
			
			/* first find the closest marker to imagePoint (within a radius EPSILON) */
			// initialization with the first point
			Anchor2D closestPoint = controlPoints.get(0);
			double closestDistance = imagePoint2D.distance(closestPoint.getPosition());
			for (int i = 1 ; i < controlPoints.size() ; i++) {
				Anchor2D pt = controlPoints.get(i);
				double dist = imagePoint2D.distance(pt.getPosition());
				if (dist < closestDistance) {
					closestDistance = dist;
					closestPoint = pt;
				}
			}
			
			// but if it is not "close enough", do nothing
			if (closestDistance > EPSILON) return;
			
			// if there is only one marker left, prevent the user to remove it
			if (detector_.getNumNuclei() < 2) {
				MessageDialog.showDialog("Cannot remove the last marker.\n"
						+ "Please add a new marker before removing this one.", 
						JOptionPane.ERROR_MESSAGE);
				return;
			}
			
			// and if it is, remove the marker
			ROI2DNucleiDetector.this.beginUpdate();
			try {
				// get the position of the click
				Point2D.Double pos = new Point2D.Double(closestPoint.getX(), 
						closestPoint.getY());
	            
	            // remove from the detector's list
				boolean removed = detector_.removeNucleus(pos);
				
				if (removed) {
					// remove the marker from the ROI
					removePoint(canvas, closestPoint);
				} else {
					// normally this should never be reached
					System.err.println("Could not remove the nucleus !");
				}
				
				hasChanged_ = true;
			} finally {
				ROI2DNucleiDetector.this.endUpdate();
			}
		}
		
	
	}

}
