/*******************************************************************************
 * Copyright (c) 2012-2013 Biomedical Image Group (BIG), EPFL, Switzerland.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Ricard Delgado-Gonzalo (ricard.delgado@gmail.com)
 *     Nicolas Chenouard (nicolas.chenouard@gmail.com)
 *     Philippe Th&#233;venaz (philippe.thevenaz@epfl.ch)
 *     Emrah Bostan (emrah.bostan@gmail.com)
 *     Ulugbek S. Kamilov (kamilov@gmail.com)
 *     Ramtin Madani (ramtin_madani@yahoo.com)
 *     Masih Nilchian (masih_n85@yahoo.com)
 *     C&#233;dric Vonesch (cedric.vonesch@epfl.ch)
 *     Virginie Uhlmann (virginie.uhlmann@epfl.ch)
 *     Cl&#233;ment Marti (clement.marti@epfl.ch)
 *     Julien Jacquemot (julien.jacquemot@epfl.ch)
 ******************************************************************************/
package plugins.big.bigsnake.snake;

import icy.util.XMLUtil;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import plugins.big.bigsnake.core.SettingsSD;
import plugins.big.bigsnake.snake.Enums.ESnakeEnergyTypeSD;
import plugins.big.bigsnake.snake.Enums.ESnakePriorShapeTypeSD;
import plugins.big.bigsnake.snake.Enums.ESnakeTargetTypeSD;
import plugins.big.bigsnake.snake.Enums.MultiresolutionTypeSD;
import plugins.big.bigsnake.snake.Enums.ShapeSpaceTypeSD;
import plugins.big.bigsnakeutils.process.process1D.BSplineBasis;
import plugins.big.bigsnakeutils.process.process1D.BSplineBasis.BSplineBasisType;

/**
 * Class that wraps the parameters of the E-Snake.
 * 
 * @version October 27, 2014
 * 
 * @author Ricard Delgado-Gonzalo (ricard.delgado@gmail.com)
 * @author: Daniel Schmitter (daniel.schmitter@epfl.ch)
 * @author Ana�s Badoual (anais.badoual@gmail.com)
 */
public class ESnakeParametersSD {

	/** Indicates the type of features to detect (bright or dark). */
	protected ESnakeTargetTypeSD detectType_ = SettingsSD.TARGET_TYPE_DEFAULT;
	/** Number of spline vector coefficients. */
	protected int M_ = SettingsSD.M_DEFAULT;
	/** Number of subdivision coefficients. */
	protected int MSub_ = SettingsSD.M_DEFAULT;
	/** Number of subdivision level. */
	protected int Level_=SettingsSD.SDLEVEL_DEFAULT;
	/** Number of m scaling factor. */
	protected int mScaling_=SettingsSD.SDMSCALING_DEFAULT;
	/** Indicates the energy function of the snake. */
	protected ESnakeEnergyTypeSD energyType_ = SettingsSD.ENERGY_TYPE_DEFAULT;
	/** Energy tradeoff factor. */
	protected double alpha_ = SettingsSD.ALPHA_DEFAULT;
	/** Prior shape information. */
	protected ESnakePriorShapeTypeSD priorShapeType_ = SettingsSD.PRIOR_SHAPE_DEFAULT;
	/** Shape space information. */
	protected ShapeSpaceTypeSD shapeSpaceType_ = SettingsSD.SHAPE_SPACE_DEFAULT;
	/** Prior-shape energy tradeoff factor. */
	protected double beta_ = SettingsSD.BETA_DEFAULT;
	/** Basis function information. */
	protected BSplineBasisType basisFunctionType_ = SettingsSD.BASIS_FUNCTION_DEFAULT;
	/** Multiresolution information. */
	protected MultiresolutionTypeSD multiresolutionType_ = SettingsSD.MULTIRESOLUTION_DEFAULT;
	/**
	 * Maximum number of iterations allowed when the <code>immortal</code> flag
	 * is <code>false</code>.
	 */
	protected int maxLife_ = SettingsSD.MAX_LIFE_DEFAULT;
	/**
	 * If <code>true</code> indicates that the snake will keep iterating till
	 * the optimizer decides so.
	 */
	protected boolean immortal_ = SettingsSD.IMMORTAL_DEFAULT;

	// ----------------------------------------------------------------------------
	// XML TAG FIELDS

	/**
	 * Label of the XML tag informing about the brightness of the target to
	 * segment (brighter or darker than the background).
	 */
	public static final String ID_DETECT_TYPE = "detectType";
	/**
	 * Label of the XML tag containing the number of control points of the
	 * snake.
	 */
	public static final String ID_M = "M";
	/**
	 * Label of the XML tag informing about the type of energy the snake uses
	 * (contour, region or mixture).
	 */
	public static final String ID_ENERGY_TYPE = "energyType";
	/**
	 * Label of the XML tag containing the energy trade-off between the contour
	 * energy and the region energy.
	 */
	public static final String ID_ALPHA = "alpha";
	/**
	 * Label of the XML tag informing about the type of prior shape that is used
	 * in the energy.
	 */
	public static final String ID_PRIOR_SHAPE_TYPE = "priorShapeType";
	/**
	 * Label of the XML tag informing about the type of shape space that is used
	 * in the energy.
	 */
	public static final String ID_SHAPE_SPACE_TYPE = "priorShapeType";
	/**
	 * Label of the XML tag informing about the type of basis function that is used.
	 */
	public static final String ID_BASIS_FUNCTION_TYPE = "basisFunctionType";
	/**
	 * Label of the XML tag containing the energy weight of the prior shape
	 * energy.
	 */
	public static final String ID_BETA = "beta";
	/** Label of the XML tag containing the maximum number of iterations. */
	public static final String ID_MAX_LIFE = "maxLife";
	/**
	 * Label of the XML tag informing if the number of iterations of the snake
	 * is limited or not.
	 */
	public static final String ID_IMMORTAL = "immortal";
	public static final String ID_MULTIRESOLUTION = "multiresolution";
	/**
	 * Label of the XML tag containing the number of subdivision
	 */
	public static final String ID_Level = "Level";
	/**

	// ============================================================================
	// PUBLIC METHODS

	/** Default constructor. */
	public ESnakeParametersSD() {
	}

	// ----------------------------------------------------------------------------

	/** Constructor. */
	public ESnakeParametersSD(ESnakeTargetTypeSD detectType, int M, int level, BSplineBasisType basisFunctionType, int m,
			ESnakeEnergyTypeSD energyType, double alpha,
			ESnakePriorShapeTypeSD priorShapeType, int maxLife,
			boolean immortal, MultiresolutionTypeSD multiresolutionType) {
		setMaxLife(maxLife);
		setM(M);
		setLevel(level);
		setBasisFunctionType(basisFunctionType);
		setMScalingFactor(m);
		setAlpha(alpha);
		setImmortal(immortal);
		setDetectType(detectType);
		setEnergyType(energyType);
		setPriorShapeType(priorShapeType);

		setMultiresolutionType(multiresolutionType);
	}
	
	// ----------------------------------------------------------------------------

	@Override
	public String toString() {
		return new String("[Snake parameters: " + ID_DETECT_TYPE + " "
				+ detectType_ + ID_M + " " + M_ + ID_Level+ " "+ Level_+ID_BASIS_FUNCTION_TYPE+" "+basisFunctionType_+ ID_ENERGY_TYPE + " "
				+ energyType_ + ID_ALPHA + " " + alpha_ + ID_PRIOR_SHAPE_TYPE
				+ " " + priorShapeType_ + ID_SHAPE_SPACE_TYPE + " " + shapeSpaceType_ + ID_BETA + " " + beta_ + ID_MAX_LIFE
				+ " " + maxLife_ + ID_IMMORTAL + " " + immortal_ +ID_MULTIRESOLUTION+" "+ "]");
	}

	// ----------------------------------------------------------------------------

	public void loadFromToXML(Node node) {

		Element detectTypeAsStringElement = XMLUtil.getElement(node,
				ID_DETECT_TYPE);
		if (detectTypeAsStringElement != null) {
			String detectTypeAsString = detectTypeAsStringElement
					.getTextContent();
			if (detectTypeAsString.compareToIgnoreCase("BRIGHT") == 0) {
				setDetectType(ESnakeTargetTypeSD.BRIGHT);
			} else if (detectTypeAsString.compareToIgnoreCase("DARK") == 0) {
				setDetectType(ESnakeTargetTypeSD.DARK);
			} else {
				setDetectType(SettingsSD.TARGET_TYPE_DEFAULT);
			}
		} else {
			setDetectType(SettingsSD.TARGET_TYPE_DEFAULT);
		}

		Element energyTypeAsStringElement = XMLUtil.getElement(node,
				ID_ENERGY_TYPE);
		if (energyTypeAsStringElement != null) {
			String energyTypeAsString = energyTypeAsStringElement
					.getTextContent();
			if (energyTypeAsString.compareToIgnoreCase("CONTOUR") == 0) {
				setEnergyType(ESnakeEnergyTypeSD.CONTOUR);
			} else if (energyTypeAsString.compareToIgnoreCase("REGION") == 0) {
				setEnergyType(ESnakeEnergyTypeSD.REGION);
			} else if (energyTypeAsString.compareToIgnoreCase("MIXTURE") == 0) {
				setEnergyType(ESnakeEnergyTypeSD.MIXTURE);
			} else {
				setEnergyType(SettingsSD.ENERGY_TYPE_DEFAULT);
			}
		} else {
			setEnergyType(SettingsSD.ENERGY_TYPE_DEFAULT);
		}

		Element priorShapeTypeAsStringElement = XMLUtil.getElement(node,
				ID_PRIOR_SHAPE_TYPE);
		if (priorShapeTypeAsStringElement != null) {
			String priorShapeTypeAsString = priorShapeTypeAsStringElement
					.getTextContent();
			if (priorShapeTypeAsString.compareToIgnoreCase("NONE") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.NONE);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("BANANA") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.BANANA);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("BOOMERANG") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.BOOMERANG);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("BRAINWHITE") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.BRAINWHITE);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("CIRCLE") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.CIRCLE);
			} else if (priorShapeTypeAsString
					.compareToIgnoreCase("CORPUSCALLOSUM") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.CORPUSCALLOSUM);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("CROSS") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.CROSS);
			} else if (priorShapeTypeAsString
					.compareToIgnoreCase("ELECTRICGUITAR") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.ELECTRICGUITAR);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("FEMUR") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.FEMUR);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("FLY") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.FLY);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("MOUSEORGAN") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.MOUSEORGAN);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("SEMICIRCLE") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.SEMICIRCLE);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("SQUARE") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.SQUARE);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("UPPERLIP") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.UPPERLIP);
			} else if (priorShapeTypeAsString.compareToIgnoreCase("CUSTOM") == 0) {
				setPriorShapeType(ESnakePriorShapeTypeSD.CUSTOM);
			} else {
				setPriorShapeType(SettingsSD.PRIOR_SHAPE_DEFAULT);
			}
		} else {
			setPriorShapeType(SettingsSD.PRIOR_SHAPE_DEFAULT);
		}
		
		Element shapeSpaceTypeAsStringElement = XMLUtil.getElement(node,
				ID_SHAPE_SPACE_TYPE);
		if (shapeSpaceTypeAsStringElement != null) {
			String shapeSpaceTypeAsString = shapeSpaceTypeAsStringElement
					.getTextContent();
			if (shapeSpaceTypeAsString.compareToIgnoreCase("SIMILARITY") == 0) {
				setShapeSpaceType(ShapeSpaceTypeSD.SIMILARITY);
			} else if (shapeSpaceTypeAsString.compareToIgnoreCase("AFFINE") == 0) {
				setShapeSpaceType(ShapeSpaceTypeSD.AFFINE);
			}
		} else {
			setShapeSpaceType(SettingsSD.SHAPE_SPACE_DEFAULT);
		}

		setMaxLife(XMLUtil.getElementIntValue(node, ID_MAX_LIFE,
				SettingsSD.MAX_LIFE_DEFAULT));
		setM(XMLUtil.getElementIntValue(node, ID_M, SettingsSD.M_DEFAULT));
		setAlpha(XMLUtil.getElementDoubleValue(node, ID_ALPHA,
				SettingsSD.ALPHA_DEFAULT));
		setBeta(XMLUtil.getElementDoubleValue(node, ID_BETA,
				SettingsSD.BETA_DEFAULT));
		setImmortal(XMLUtil.getElementBooleanValue(node, ID_IMMORTAL,
				SettingsSD.IMMORTAL_DEFAULT));
		setLevel(XMLUtil.getElementIntValue(node, ID_Level, SettingsSD.SDLEVEL_DEFAULT));
	}

	// ----------------------------------------------------------------------------

	public void saveToXML(Element node) {
		switch (detectType_) {
		case BRIGHT:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_DETECT_TYPE,
					"BRIGHT");
			break;
		case DARK:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_DETECT_TYPE,
					"DARK");
			break;
		}
		XMLUtil.setElementIntValue(node, ESnakeParametersSD.ID_M, M_);
		XMLUtil.setElementIntValue(node, ESnakeParametersSD.ID_Level, Level_);
		switch (energyType_) {
		case CONTOUR:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_ENERGY_TYPE,
					"CONTOUR");
			break;
		case REGION:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_ENERGY_TYPE,
					"REGION");
			break;
		case MIXTURE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_ENERGY_TYPE,
					"MIXTURE");
			break;
		}
		XMLUtil.setElementDoubleValue(node, ESnakeParametersSD.ID_ALPHA, alpha_);
		switch (priorShapeType_) {
		case NONE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"NONE");
			break;
		case BANANA:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"BANANA");
			break;
		case BOOMERANG:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"BOOMERANG");
			break;
		case BRAINWHITE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"BRAINWHITE");
			break;
		case CIRCLE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"CIRCLE");
			break;
		case CORPUSCALLOSUM:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"CORPUSCALLOSUM");
			break;
		case CROSS:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"CROSS");
			break;
		case ELECTRICGUITAR:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"ELECTRICGUITAR");
			break;
		case FEMUR:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"FEMUR");
			break;
		case FLY:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"FLY");
			break;
		case MOUSEORGAN:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"MOUSEORGAN");
			break;
		case SEMICIRCLE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"SEMICIRCLE");
			break;
		case SQUARE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"SQUARE");
			break;
		case UPPERLIP:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"UPPERLIP");
			break;
		case CUSTOM:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_PRIOR_SHAPE_TYPE,
					"CUSTOM");
			break;
		default:
			break;
		}
		switch (shapeSpaceType_) {
		case SIMILARITY:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_SHAPE_SPACE_TYPE,
					"SIMILARITY");
			break;
		case AFFINE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_SHAPE_SPACE_TYPE,
					"AFFINE");
			break;
		default:
			break;
		}
		switch (basisFunctionType_) {
		case QUADRATICBSPLINE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_BASIS_FUNCTION_TYPE,
					"QUADRATICBSPLINE");
			break;
		case CUBICBSPLINE:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_BASIS_FUNCTION_TYPE,
					"CUBICBSPLINE");
			break;
		case ESPLINE3:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_BASIS_FUNCTION_TYPE,
					"ESPLINE3");
			break;
		case DESLAURIERSDUBUC:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_BASIS_FUNCTION_TYPE,
					"DESLAURIERSDUBUC");
			break;
		case GDESLAURIERSDUBUC:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_BASIS_FUNCTION_TYPE,
					"GDESLAURIERSDUBUC");
			break;
		case GDESLAURIERSDUBUC2:
			XMLUtil.setElementValue(node, ESnakeParametersSD.ID_BASIS_FUNCTION_TYPE,
					"GDESLAURIERSDUBUC2");
			break;
		default:
			break;
		}
		XMLUtil.setElementDoubleValue(node, ESnakeParametersSD.ID_BETA, beta_);
		XMLUtil.setElementDoubleValue(node, ESnakeParametersSD.ID_MAX_LIFE,
				maxLife_);
		XMLUtil.setElementBooleanValue(node, ESnakeParametersSD.ID_IMMORTAL,
				immortal_);
	}

	// ============================================================================
	// GETTERS AND SETTERS

	public ESnakeTargetTypeSD getDetectType() {
		return detectType_;
	}

	// ----------------------------------------------------------------------------

	public void setDetectType(ESnakeTargetTypeSD detectType) {
		detectType_ = detectType;
	}

	// ----------------------------------------------------------------------------

	public int getM() {
		return M_;
	}
	
	// ----------------------------------------------------------------------------

	public int getMSub() {
		return MSub_;
	}

	// ----------------------------------------------------------------------------

	public int getSubdivisionLevel() {
		return Level_;
	}

	// ----------------------------------------------------------------------------
	
	public int getMScalingFactor() {
		return mScaling_;
	}

	// ----------------------------------------------------------------------------
	public void setM(int M) {
		M_ = M;
	}

	// ----------------------------------------------------------------------------

	public void setLevel(int level) {
		Level_ = level;
	}
	
	// ----------------------------------------------------------------------------

	public void setBasisfunctionType(BSplineBasisType basisFunctionType) {
		basisFunctionType_ = basisFunctionType;
	}

	// ----------------------------------------------------------------------------
	
	public void setMScalingFactor(int m) {
		mScaling_ = m;
	}

	// ----------------------------------------------------------------------------
	public ESnakeEnergyTypeSD getEnergyType() {
		return energyType_;
	}

	// ----------------------------------------------------------------------------

	public void setEnergyType(ESnakeEnergyTypeSD energyType) {
		energyType_ = energyType;
	}

	// ----------------------------------------------------------------------------

	public double getAlpha() {
		return alpha_;
	}

	// ----------------------------------------------------------------------------

	public void setAlpha(double alpha) {
		alpha_ = alpha;
	}

	// ----------------------------------------------------------------------------

	public ESnakePriorShapeTypeSD getPriorShapeType() {
		return priorShapeType_;
	}

	// ----------------------------------------------------------------------------

	public void setPriorShapeType(ESnakePriorShapeTypeSD priorShapeType) {
		priorShapeType_ = priorShapeType;
	}
	
	// ----------------------------------------------------------------------------

	public ShapeSpaceTypeSD getShapeSpaceType() {
		return shapeSpaceType_;
	}
	
	// ----------------------------------------------------------------------------

	public MultiresolutionTypeSD getMultiresolutionType() {
		return multiresolutionType_;
	}
	
	// ----------------------------------------------------------------------------

	public BSplineBasisType getBasisFunctionType() {
		return basisFunctionType_;
	}
	
	// ----------------------------------------------------------------------------

	public void setShapeSpaceType(ShapeSpaceTypeSD shapeSpaceType) {
		shapeSpaceType_ = shapeSpaceType;
	}
	
	// ----------------------------------------------------------------------------

	public void setBasisFunctionType(BSplineBasisType basisFunctionType) {
		basisFunctionType_ = basisFunctionType;
	}


	// ----------------------------------------------------------------------------

	public double getBeta() {
		return beta_;
	}

	// ----------------------------------------------------------------------------

	public void setBeta(double beta) {
		beta_ = beta;
	}
	
	// ----------------------------------------------------------------------------

	public void setMultiresolutionType(MultiresolutionTypeSD multiresolutionType) {
		multiresolutionType_ = multiresolutionType;
	}

	// ----------------------------------------------------------------------------

	public int getMaxLife() {
		return maxLife_;
	}

	// ----------------------------------------------------------------------------

	public void setMaxLife(int maxLife) {
		maxLife_ = maxLife;
	}

	// ----------------------------------------------------------------------------

	public boolean isImmortal() {
		return immortal_;
	}

	// ----------------------------------------------------------------------------

	public void setImmortal(boolean immortal) {
		immortal_ = immortal;
	}
}
