package plugins.big.bigsnake.snake;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import kn.uni.voronoitreemap.j2d.PolygonSimple;
import plugins.big.bigsnake.snake.Enums.GlobalSnakeSortingType;
import plugins.big.bigsnake.snake.Enums.GlobalSnakeInitialization;
import plugins.big.bigsnake.voronoi.VoronoiDiagram;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DNode;

/**
 * Object that manufactures GlobalSnake initial control points,
 * for now only from a Voronoi diagram
 * (aims to be extended to other GlobalSnake initialization methods).
 * 
 * @version January 13, 2018
 * 
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 */
public class GlobalSnakeInitializer {

	private VoronoiDiagram voronoi_ = null;
	
	private List<Point2D.Double> seeds_ = null;
	private List<SubsnakeInitialShape> shapes_ = null;
	
	private GlobalSnakeInitialization initType_ = null;
	private GlobalSnakeSortingType sortingType_ = null;
	
	private final boolean showDiagram_ = false; // debug only
	
	// =====================
	// PUBLIC METHODS
		
	public GlobalSnakeInitializer(GlobalSnakeInitialization initType,
			GlobalSnakeSortingType sortingType) {
		initType_ = initType;
		sortingType_ = sortingType;
	}
	
	public GlobalSnakeInitializer(GlobalSnakeInitialization initType,
			GlobalSnakeSortingType sortingType, VoronoiDiagram voronoi) {
		this(initType, sortingType);
		setVoronoiDiagram(voronoi);
	}
	
	
	public void setSpecificSeeds(List<Point2D.Double> points) {
		if (initType_ != GlobalSnakeInitialization.TWO_CELLS) {
			System.err.println("Manual supply of seeds only work for 2 seeds");
			return;
		}
		seeds_ = points;
	}
	
	
	public void setVoronoiDiagram(VoronoiDiagram voronoi) {
		if (initType_ != GlobalSnakeInitialization.BOUNDED_VORONOI) {
			System.err.println(String.format("Can't set Voronoi diagram "
					+ "with init type %s", initType_));
			return;
		}
		voronoi_ = voronoi;
	}
	
	
	public List<SubsnakeInitialShape> getInitialShapes() {
		
		shapes_ = new ArrayList<SubsnakeInitialShape>();
		
		if (seeds_ == null) {
			seeds_ = new ArrayList<Point2D.Double>();
		}
		
		switch(initType_) {
		case BOUNDED_VORONOI:
			computeShapesFromVoronoi();
			break;
		case TWO_CELLS:
			computeShapesFromTwoSeeds();
			break;
		}
		
		return shapes_ ;
	}
	
	
	public List<Point2D.Double> getOrderedSeeds() {
		if (sortingType_ != GlobalSnakeSortingType.NONE) {
			// apply the new ordering to the seeds
			updateSeeds();
		}
		return seeds_;
	}
	
	
	private void computeShapesFromVoronoi() {
		
		if (!voronoi_.isComputed()) voronoi_.runComputation(showDiagram_);
		
		// compute Subsnake shapes
		buildVoronoiShapes();
		
		/* in order to ensure that each initialization from same nuclei is the same,
		 * the array is sorted as function of either subsnakes centroids or areas
		 */
		sortShapes();
	}
	
//	private void sortNodes() {
//		
//		for (int i = 0 ; i < shapes_.size() ; i++) {
//			
//		}
//	}

	private void sortShapes() {
		Comparator<SubsnakeInitialShape> comparator = null;
		switch (sortingType_) {
		case SUBSNAKE_CENTROID_X:
			comparator = Comparator.comparing(SubsnakeInitialShape::getCentroidX)
					.thenComparing(SubsnakeInitialShape::getCentroidY);
			break;
		case SUBSNAKE_CENTROID_Y:
			comparator = Comparator.comparing(SubsnakeInitialShape::getCentroidY)
					.thenComparing(SubsnakeInitialShape::getCentroidX);
			break;
		case SUBSNAKE_AREA_INC:
			comparator = Comparator.comparing(SubsnakeInitialShape::getArea);
			break;
		case SUBSNAKE_AREA_DEC:
			comparator = Comparator.comparing(SubsnakeInitialShape::getArea).reversed();
			break;
		case NUCLEI_CENTROID_X:
			comparator = Comparator.comparing(SubsnakeInitialShape::getStartingPointX)
					.thenComparing(SubsnakeInitialShape::getStartingPointY);
			break;
		case NUCLEI_CENTROID_Y:
			comparator = Comparator.comparing(SubsnakeInitialShape::getStartingPointY)
					.thenComparing(SubsnakeInitialShape::getStartingPointX);
			break;
		case NONE:
			comparator = null;
			break;
		}
		
		if (sortingType_ != GlobalSnakeSortingType.NONE) {
			shapes_.sort(comparator);
		}
	}

	private void computeShapesFromTwoSeeds() {
		double x1 = seeds_.get(0).x;
		double y1 = seeds_.get(0).y;
		double x2 = seeds_.get(1).x;
		double y2 = seeds_.get(1).y;
		
		final double diffX = Math.abs(x1 - x2);
		final double diffY = Math.abs(y1 - y2);
		final boolean horizontal = (diffY == 0);
		final boolean vertical = (diffX == 0);
		
		if (vertical && horizontal) {
			System.err.println("Seeds can't have same coordinates");
			return;
		}
		else if (horizontal) {
			shapes_ = horizontallyAlignedSeeds(x1, x2, y1, y2);
		}
		else if (vertical) {
			shapes_ = verticallyAlignedSeeds(x1, x2, y1, y2);
		}
		else {
			System.err.println("2-cells initilization not supported, "
					+ "except if seeds are (perfectly) horizontally or vertically aligned");
			return;
		}
	}
	
	
	private void buildVoronoiShapes() {
		PolygonSimple[] diagramPolygons = voronoi_.getPolygons();
		Point2D.Double[] diagramSeeds = voronoi_.getSitesAsPoints();
		
		for (int i = 0 ; i < diagramPolygons.length ; i++) {
			
			Point2D.Double seedPoint = diagramSeeds[i];
			PolygonSimple polygon = diagramPolygons[i];
			int numCoefs = polygon.getNumPoints();
			
			SubsnakeInitialShape subsnakeCoef = 
					new SubsnakeInitialShape(seedPoint);
			
			double[] coefX = polygon.getXPoints();
			double[] coefY = polygon.getYPoints();
			
			for (int j = 0 ; j < numCoefs ; j++) {
				Snake2DNode node = new Snake2DNode(
						Math.round(coefX[j]), 
						Math.round(coefY[j]));
				subsnakeCoef.add(node);
			}
			
			/* Sort nodes inside the Subsnake,
			 * to avoid randomness that may come from the Voronoi diagram algorithm */
			Comparator<Snake2DNode> comparator = Comparator.comparing(Snake2DNode::getX)
					.thenComparing(Snake2DNode::getY);
			subsnakeCoef.sortAndPreserveShape(comparator);
			
			shapes_.add(subsnakeCoef);
			seeds_.add(seedPoint);
		}
	}

	
	private void updateSeeds() {
		// reinitialize seeds_ so it is consistent with shapes_'s starting points
		seeds_ = new ArrayList<Point2D.Double>();
		for (SubsnakeInitialShape shape : shapes_) {
			seeds_.add(shape.getStartingPoint());
		}
	}
	
	
	private List<SubsnakeInitialShape> verticallyAlignedSeeds(double x1, double x2, double y1, double y2) {
		final double diffY = Math.abs(y1 - y2);
		
		List<SubsnakeInitialShape> allCoefs = new ArrayList<SubsnakeInitialShape>();
		
		// prepare all nodes
		final Snake2DNode upperMostNode = 
				new Snake2DNode(Math.round(x1), Math.round(y1-diffY/3));
		final Snake2DNode centralLeftNode  = 
				new Snake2DNode(Math.round(x1-diffY/2), Math.round(y2-diffY/2));
		final Snake2DNode centralRightNode = 
				new Snake2DNode(Math.round(x1+diffY/2), Math.round(y2-diffY/2));
		final Snake2DNode lowerMostNode = 
				new Snake2DNode(Math.round(x2), Math.round(y2+diffY/3));
		
		// First Subsnake
		final Point2D.Double startPt1 = new Point2D.Double(x1, y1);
		SubsnakeInitialShape subsnake1Coef = new SubsnakeInitialShape(startPt1);
		subsnake1Coef.add(upperMostNode);
		subsnake1Coef.add(centralLeftNode);
		subsnake1Coef.add(centralRightNode);
		allCoefs.add(subsnake1Coef);
		
		// Second Subsnake, having two common nodes with subsnake above
		final Point2D.Double startPt2 = new Point2D.Double(x2, y2);
		SubsnakeInitialShape subsnake2Coef = new SubsnakeInitialShape(startPt2);
		subsnake2Coef.add(lowerMostNode);
		subsnake2Coef.add(centralRightNode);
		subsnake2Coef.add(centralLeftNode);
		allCoefs.add(subsnake2Coef);
		
		return allCoefs;
}


	private List<SubsnakeInitialShape> horizontallyAlignedSeeds(double x1, double x2, double y1, double y2) {
		final double diffX = Math.abs(x1 - x2);
		
		List<SubsnakeInitialShape> allCoefs = new ArrayList<SubsnakeInitialShape>();
		
		// prepare all nodes
		Snake2DNode leftMostNode = new Snake2DNode(Math.round(x1-diffX/3), Math.round(y1));
		Snake2DNode centralUpperNode = new Snake2DNode(Math.round(x2-diffX/2), Math.round(y1-diffX/2));
		Snake2DNode centralLowerNode = new Snake2DNode(Math.round(x2-diffX/2), Math.round(y1+diffX/2));
		Snake2DNode rightMostNode = new Snake2DNode(Math.round(x2+diffX/3), Math.round(y2));
		
		// First Subsnake
		SubsnakeInitialShape subsnake1Coef = new SubsnakeInitialShape();
		subsnake1Coef.add(leftMostNode);
		subsnake1Coef.add(centralUpperNode);
		subsnake1Coef.add(centralLowerNode);
		allCoefs.add(subsnake1Coef);
		
		// Second Subsnake, having two common nodes with subsnake above
		SubsnakeInitialShape subsnake2Coef = new SubsnakeInitialShape();
		subsnake2Coef.add(rightMostNode);
		subsnake2Coef.add(centralLowerNode);
		subsnake2Coef.add(centralUpperNode);
		allCoefs.add(subsnake2Coef);
		
		return allCoefs;
	}

}
