package plugins.big.bigsnake.snake;

import java.awt.Color;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import plugins.big.bigsnake.core.ImageLUTContainerSD;
import plugins.big.bigsnake.graph.GlobalSnakeGraph;
import plugins.big.bigsnake.snake.Enums.ESnakeEnergyTypeSD;
import plugins.big.bigsnake.snake.Enums.ESnakeTargetTypeSD;
import plugins.big.bigsnake.snake.Enums.GlobalSnakeEdgeHandlingType;
import plugins.big.bigsnake.snake.Enums.GlobalSnakeEnergyTypeSD;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DNode;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DScale;
import plugins.big.bigsnakeutils.process.process1D.BSplineBasis;
import plugins.big.bigsnakeutils.process.process1D.BSplineBasis.BSplineBasisType;


/**
 * GlobalSnake for jointly segmentation
 * (based on Exponential spline snake).
 * 
 * @version January 13, 2018
 * 
 * @author Ana�s Badoual (anais.badoual@gmail.com)
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 */
public class GlobalSnakeSD extends ESnakeSD {
	
	/** Number of <code>SubsnakeSD</code>. */
	private int nSubsnakes_ = 1;
	
	/** Total number of nodes in all <code>SubsnakeSD</code>. */
	private int totalM_ = 0;
	private int totalMSub_ = 0;
	
	private int[] allM_ = null;
	private int[] allMSub_ = null;
	private int globalLevel_ = 1;
	private List<SubsnakeInitialShape> allSubsnakeCoefs_ = null;
	
	/**
	 * List of all <code>SubsnakeSD</code>.
	 */
	private SubsnakeSD[] subsnakes_ = null;
	
	/** All parameters. */
	private ESnakeParametersSD[] allSubsnakeParameters_ = null;
	
	/**
	 * Samples of the x coordinates of the global snake contour obtained by
	 * subdivision.
	 */
//	private double[] xGlobalSnakeContourSubdivision_ = null;
	/**
	 * Samples of the y coordinates of the global snake contour obtained by
	 * subdivision.
	 */
//	private double[] yGlobalSnakeContourSubdivision_ = null;
	
	/**
	 * Exactly same sized array than subsnakes_,
	 * filled with position (index) of subsnakes_ nodes inside nodes_
	 */
	private int[] subsnakesToSnakeCoefMap_ = null;
	private int[] subsnakesToSnakeSubdivCoefMap_ = null;
	
	/**
	 * Binary arrays to keep track of control 
	 * and subdivision nodes from Subsnakes in GlobalSnake.
	 */
	private boolean[] coefToTrack_ = null;
	private boolean[] subdivCoefToTrack_ = null;
	private boolean coefUpToDate_ = false;
	private boolean subdivCoefUpToDate_ = false;
	private boolean[] coefToHide_ = null;
	private boolean[] coefToFreeze_ = null;
	
	/**
	 * Mapping between old (= original) control nodes
	 * and new (= after intermediate subdivision) ones,
	 * so when a settings has been adjusted, 
	 * control nodes are updated to current positions.
	 */
	private List<int[]> oldToNewCoefsMap_ = null;
	
	/** Select the way to compute the global snake total energy. */
	private GlobalSnakeEnergyTypeSD globalEnergyType_ = null;
	
	/** Remove or not the 'common edge', 
	 *  i.e. those that are doubled between two given nodes.
	 */
	private boolean showCommonEdges_ = false;
	
	/**
	 * Graph defined by GlobalSnake control nodes.
	 */
	private GlobalSnakeGraph graph_ = null;
	
	/**
	 * Intermediate subdivision
	 */
	private int intermediateSubdiv_ = 0;
	
	/**
	 * Treatment of control nodes that belong to a common edge :
	 * - either freeze it (in the sense of a Snake2DNode, basically the optimizer does not move it)
	 * - or link it to the position of its counter part on the 'kept' common edge
	 */
	private final GlobalSnakeEdgeHandlingType controlNodeHandling_ = 
			GlobalSnakeEdgeHandlingType.HIDE_AND_FREEZE;
	
	/**
	 * Flag that tells if the global snake has been initialized from scratch, 
	 * or updated from new settings
	 */
	private boolean updated_ = false;

	
	// ============================================================================
	// PUBLIC METHODS
	
	/** Constructor */
	public GlobalSnakeSD(ImageLUTContainerSD imageLUTs, 
			GlobalSnakeParametersSD parameters) {
		
		super(imageLUTs, parameters, null, null);
		
		/* colors are controlled by Subsnakes, but it can be modified
		 * (see ROI2DSnakeSD and overriden methods on its children classes 
		 * ROI2DSsubsnakeCurveSD / ROI2DGlobalSnakeSD)
		 */
		SNAKE_CONTOUR_COLOR = new Color(255, 0, 0, 255);  // low opacity for smooth curve
		CONTROL_POLYGON_COLOR = new Color(255, 255, 0, 0);  // opacity to 0 for polygons
		
		nSubsnakes_ = parameters.getNSubsnakes();
		allSubsnakeCoefs_ = parameters.getAllSubsnakeCoefs();
		allSubsnakeParameters_ = parameters.getAllSubsnakeParameters();
		
		globalLevel_ = parameters.getGlobalLevel();
		
		intermediateSubdiv_ = parameters.getIntermediateSubdiv();
		showCommonEdges_ = parameters.showCommonEdges();
		
		energyType_ = parameters.getEnergyType();
		globalEnergyType_ = parameters.getGlobalEnergyType();
		
		coefUpToDate_ = false;
		subdivCoefUpToDate_ = false;
		updated_ = false;
		
		initialize();
	}
	
	
	@Override
	public String toString() {
		String eol = System.getProperty("line.separator");
		String title = "GlobalSnake : ";
		String lineStruct = "nSubsnakes_ = " + nSubsnakes_ + 
				", M_ = " + M_ + ", MSub_ = " + MSub_ + ", ";
		String lineLevels = "global subdiv level = " + globalLevel_ + 
				", effective subdiv level = " + Level_ + ", ";
		String lineLife;
		if (immortal_) {
			lineLife = "immortal";
		} else {
			lineLife = "life = " + life_;
		}
        return title + lineStruct + eol + lineLevels + lineLife;
    }
	
	
	public SubsnakeSD getSubsnake(int subsnakeId) {
		return subsnakes_[subsnakeId];
	}
	
	
	public int getNumSubsnakes() {
		return nSubsnakes_;
	}
	
	
	public int getGlobalLevel() {
		return globalLevel_;
	}
	
	
	@Override
	public Snake2DNode[] getNodes() {
		buildControlNodes(false);
		return coef_;
	}
	
	
	@Override
	public Snake2DNode[] getSubdivisionPoints() {
		buildSubdivisionNodes(false);
		return subdivCoef_;
	}
	
	
	/**
	 * This mutator does not behave like the original method from ESnakeSD !
	 */
	@Override
	public void setNodes(Snake2DNode[] nodes) {
		coefUpToDate_ = false;
		subdivCoefUpToDate_ = false;
		
		int mapIdx  = 0;  // for loop over control nodes
		int mapIdxS = 0;  // for loop over subdivision nodes
		
		for (int i = 0 ; i < nSubsnakes_ ; i++) {
			
			int targetNumNodes = allM_[i];
			Snake2DNode[] newSubsnakeControlNodes  = new Snake2DNode[targetNumNodes];
			
			for (int j = 0 ; j < targetNumNodes ; j++) {
				newSubsnakeControlNodes[j]  = nodes[subsnakesToSnakeCoefMap_[mapIdx]];
				mapIdx++;
			}
			
			/* Set control nodes and recompute subdivision points
			 * (chain call to updateSkin() of that subsnake)
			 */
			subsnakes_[i].setNodes(newSubsnakeControlNodes);

			// then set the 'hidden' state to some subdivision nodes that belong to common edges
			Snake2DNode[] newSubdivNodes = subsnakes_[i].getSubdivisionPoints();
			for (int j = 0 ; j < allMSub_[i] ; j++) {
				/* Hide the subdivision node only if this is not a coef that 
				 * is not tracked by the global snake AND that is not also a control node
				 */
				boolean isNotAlsoControlNode = !(Arrays.asList(coef_)
						.contains(newSubdivNodes[j]));
				boolean isNotToTrack = !(subdivCoefToTrack_[mapIdxS]);
				
				if (isNotToTrack && isNotAlsoControlNode) {
					newSubdivNodes[j].hide();
				}
				mapIdxS++;
			}
			subsnakes_[i].saveHiddenStates();
		}
		
		// now that every subsnake's skin is updated, rebuild container nodes and controlNodes
		buildControlNodes(false);
		buildSubdivisionNodes(false);
		
		updateSnakeSkin();
		updateAreaSubdivision();
		
		validateSnake();
	}
	
	
	public void setGlobalLevelSubdivision(int level) {
		globalLevel_ = level;
	}
	
	
	@Override
	public void setLevelSubdivision(int level) {
		super.setLevelSubdivision(level);
		for (int i = 0 ; i < nSubsnakes_ ; i++)
			allSubsnakeParameters_[i].setLevel(level);
	}
	
	
	public void equalizeLevelSubdivision(int level) {
		setLevelSubdivision(level);
		setGlobalLevelSubdivision(level);
	}
	
	
	public void setEnergyType(ESnakeEnergyTypeSD energyType) {
		energyType_ = energyType;
		for (int i = 0 ; i < nSubsnakes_ ; i++)
			subsnakes_[i].setEnergyType(energyType);
	}
	
	
	public void setMaxLife(int maxLife) {
		maxLife_ = maxLife;
		for (int i = 0 ; i < nSubsnakes_ ; i++)
			subsnakes_[i].setMaxLife(maxLife);
	}
	
	
	public void setImmortal(boolean immortal) {
		immortal_ = immortal;
		for (int i = 0 ; i < nSubsnakes_ ; i++)
			subsnakes_[i].setImmortal(immortal);
	}
	
	
	public void setDetectType(ESnakeTargetTypeSD detectType) {
		detectType_ = detectType;
		for (int i = 0 ; i < nSubsnakes_ ; i++)
			subsnakes_[i].setDetectType(detectType);
	}
	
	
	public void setBasisFunctionType(BSplineBasisType basisFunctionType) {
		basisFunctionType_ = basisFunctionType;
		BASIS_FUNCTION = basisFunctionType;
		for (int i = 0 ; i < nSubsnakes_ ; i++)
			subsnakes_[i].setBasisFunctionType(basisFunctionType);
	}
	
	
	@Override
	public double energy() {
		double energy;
		switch (globalEnergyType_) {
		case REAL:
			energy = super.energy();
			for (SubsnakeSD ss : subsnakes_)
				ss.life_--;
			break;
		case SUM:
			energy = sumEnergy();
			break;
		default:
			energy = sumEnergy();
			break;
		}
		return energy;
	}


	@Override
	public void subdivision() {
//		buildSubdivisionNodes(false);
		System.err.println("Basis function type not supported yet in GlobalSnake.");
	}
	
	
	@Override
	public void subdivisionDeslaurierDubuc() {
		buildSubdivisionNodes(false);
		computeDerivativeSubdivPointsDeslaurierDubuc();
	}
	
	
	@Override
	public void subdivisionNSDeslaurierDubuc() {
//		buildSubdivisionNodes(false);
		System.err.println("Basis function type not supported yet in GlobalSnake.");
	}
	
	
	@Override
	public void subdivisionNSDeslaurierDubuc2() {
//		buildSubdivisionNodes(false);
		System.err.println("Basis function type not supported yet in GlobalSnake.");
	}
	
	
	@Override
	public void nonStationnarySubdivision() {
//		buildSubdivisionNodes(false);
		System.err.println("Basis function type not supported yet in GlobalSnake.");
	}
	
	
	@Override
	public Snake2DScale[] getScales() {
		int numScales = getNumScales();
		Snake2DScale[] skin = new Snake2DScale[numScales];

		double[] xpoints = new double[M_];
		double[] ypoints = new double[M_];
		for (int k = 0; k < M_; k++) {
			if (!coef_[k].isHidden()) {
				xpoints[k] = coef_[k].x;
				ypoints[k] = coef_[k].y;
			}
		}
		skin[0] = new Snake2DScale(xSnakeContourSubdivision_, ySnakeContourSubdivision_, 
				xSnakeContourSubdivision_.length, SNAKE_CONTOUR_COLOR, false);
		
		skin[1] = new Snake2DScale(xpoints, ypoints, 
				M_, CONTROL_POLYGON_COLOR, false);

		return skin;
	}
	
	
	public ImageLUTContainerSD getImageLUTs() {
		return imageLUTs_;
	}
	
	
	@Override
	public ESnakeParametersSD getSnakeParameters() {
		refreshInitCoefs();  // retrieve the current state of control nodes
		return new GlobalSnakeParametersSD(allSubsnakeCoefs_, 
				globalLevel_, detectType_, basisFunctionType_, energyType_,
				globalEnergyType_, maxLife_, immortal_, showCommonEdges_,
				intermediateSubdiv_);
	}
	
	
	@Override
	public void setSnakeParameters(ESnakeParametersSD parameters) {
		boolean reinitialize;
		refreshInitCoefs();  // retrieve the current state of control nodes
		
		int maxLife = parameters.getMaxLife();
		boolean immortal = parameters.isImmortal();
		ESnakeTargetTypeSD detectType = parameters.getDetectType();
		ESnakeEnergyTypeSD energyType = parameters.getEnergyType();
		GlobalSnakeEnergyTypeSD globalEnergyType = 
				((GlobalSnakeParametersSD) parameters).getGlobalEnergyType();
		BSplineBasisType basisFunctionType = parameters.getBasisFunctionType();
		int level = parameters.getSubdivisionLevel();
		List<SubsnakeInitialShape> allSubsnakeCoefs = 
				((GlobalSnakeParametersSD) parameters).getAllSubsnakeCoefs();
		ESnakeParametersSD[] allSubsnakeParameters =
				((GlobalSnakeParametersSD) parameters).getAllSubsnakeParameters();
		boolean showCommonEdges = 
				((GlobalSnakeParametersSD) parameters).showCommonEdges();
		int intermediateSubdiv = 
				((GlobalSnakeParametersSD) parameters).getIntermediateSubdiv();
		
		if (maxLife != maxLife_) {
			setMaxLife(maxLife);
			reinitialize = false;
		} else if (immortal != immortal_) {
			setImmortal(immortal);
			reinitialize = false;
		} else if (detectType != detectType_) {
			setDetectType(detectType);
			reinitialize = false;
		} else if (energyType != energyType_) {
			setEnergyType(energyType);
			reinitialize = false;
		} else if (globalEnergyType != globalEnergyType_) {
			globalEnergyType_ = globalEnergyType;
			reinitialize = false;
		} else if (allSubsnakeCoefs.size() != nSubsnakes_) {
			nSubsnakes_ = allSubsnakeCoefs.size();
			reinitialize = true;
		} else if (basisFunctionType != basisFunctionType_) {
			setBasisFunctionType(basisFunctionType);
			reinitialize = true;
		} else if (showCommonEdges != showCommonEdges_) {
			showCommonEdges_ = showCommonEdges;
			setLevelSubdivision(level);
			reinitialize = true;
		} else if (intermediateSubdiv != intermediateSubdiv_) {
			intermediateSubdiv_ = intermediateSubdiv;
			setLevelSubdivision(level);
			reinitialize = true;
		} else if (level != globalLevel_) {
			equalizeLevelSubdivision(level);
			reinitialize = true;
		} else {
			/* in case none of the above cases were reached...
			 * typically when the convex hull dilation coef has changed (= allSubsnakeCoefs changed)
			 */
			nSubsnakes_ = allSubsnakeCoefs.size();
			allSubsnakeCoefs_ = allSubsnakeCoefs;
			allSubsnakeParameters_ = allSubsnakeParameters;
			reinitialize = true;
		}

		// if the change requires it, re-initialize the globalSnake
		if (reinitialize) {
			coefUpToDate_ = false;
			subdivCoefUpToDate_ = false;
			initialize();
		}
		updated_ = true;
		
		reviveSnake();
	}


	@Override
	public void initialize(boolean initShape) {
		// do nothing (see private version of initialize() below)
	}
	
	
	/**
	 * TODO : add documentation
	 */
	public Snake2DNode getCorrectSubdivisionNode(int subsnakeId, int subdivisionNodeIndex) {
		
		int[][] map = graph_.getSubdivisionMappingForSubsnake(subsnakeId);
		
		int correctSubsnakeId = map[subdivisionNodeIndex][0];
		int correctSubdivisionNodeIndex = map[subdivisionNodeIndex][1];
		
		boolean isSameNode = (correctSubsnakeId == subsnakeId) && 
				(correctSubdivisionNodeIndex == subdivisionNodeIndex);
		if (isSameNode) {
			return null;
		}
		
		int idx1 = correctSubsnakeId;
		int idx2 = correctSubdivisionNodeIndex;
		Snake2DNode correctNode = subsnakes_[idx1].getSubdivisionPoints()[idx2];
		
		return correctNode;
	}
	
	
//	@Override
//	protected void updateSnakeSkin() {
//		super.updateSnakeSkin();
//		xGlobalSnakeContourSubdivision_ = new double[MSub_];
//		yGlobalSnakeContourSubdivision_ = new double[MSub_];
//		for (int i = 0 ; i < MSub_ ; i++) {
//			xGlobalSnakeContourSubdivision_[i] = subdivCoef_[i].x;
//			yGlobalSnakeContourSubdivision_[i] = subdivCoef_[i].y;
//		}
//	}
	
	
	/**
	 * The whole area is approximated 
	 * by the sum of subsnakes areas 
	 * TODO : may be temporary, but it works
	 */
	@Override
	protected void updateAreaSubdivision() {
		double area = 0;
		for (int i = 0 ; i < nSubsnakes_ ; i++) {
			subsnakes_[i].updateAreaSubdivision();
			area += subsnakes_[i].getArea();
		}
		area_ = area;
	}

	
	// ============================================================================
	// PRIVATE METHODS
	
	/** 
	 * Construct the lists of subdivision nodes of all subsnakes.
	 */
	private void buildSubdivisionNodes(boolean initNodes) {
		if (subdivCoefUpToDate_) return;
		
		if (initNodes) {
			subsnakesToSnakeSubdivCoefMap_ = new int[totalMSub_];
			subdivCoefToTrack_ = new boolean[totalMSub_];
		}
		
		int backupMSub = MSub_;
		
		ArrayList<Snake2DNode> tempSubdivCoef = new ArrayList<Snake2DNode>();
		int mapIdx = 0;
		for (int i = 0 ; i < nSubsnakes_ ; i++) {
			
			Snake2DNode[] currentSubsnakeNodes = subsnakes_[i].getSubdivisionPoints();
				
			for (int j = 0 ; j < allMSub_[i] ; j++) {
				
				Snake2DNode currentNode = currentSubsnakeNodes[j];
				
				/* if initialization step */
				if (initNodes) {
					int currentNodePosition = tempSubdivCoef.indexOf(currentNode);
					boolean isNewNode = (currentNodePosition == -1);
					
					boolean notInCommonEdge = true;
					if (!showCommonEdges_) {
						notInCommonEdge = belongToCleanEdge(currentNode);
					}
					
					if (isNewNode && notInCommonEdge) {
						tempSubdivCoef.add(currentNode);
						int newNodePosition = tempSubdivCoef.size() - 1;
						subsnakesToSnakeSubdivCoefMap_[mapIdx] = newNodePosition;
						subdivCoefToTrack_[mapIdx] = true;
					} else {
						subsnakesToSnakeSubdivCoefMap_[mapIdx] = currentNodePosition;
						subdivCoefToTrack_[mapIdx] = false;
					}
				}
				/* if not initialization step */
				else {
					if (subdivCoefToTrack_[mapIdx]) {
						tempSubdivCoef.add(currentNode);
					}
				}
				
				mapIdx++;
			}
		}
		
		MSub_ = tempSubdivCoef.size();
		if ((!initNodes) && (MSub_ != backupMSub)) {
			System.err.println("Warning : GlobalSnake.MSub_ has changed.");
		}
		
		subdivCoef_ = new Snake2DNode[MSub_];
		tempSubdivCoef.toArray(subdivCoef_); // update corresponding parent class attribute
	
		subdivCoefUpToDate_ = true;
	}

	
	/** 
	 * Construct the lists of control nodes of all subsnakes.
	 */
	private void buildControlNodesMain(boolean initNodes, boolean withIntermedSubdiv) {
		if (coefUpToDate_) return;
		
		if (initNodes) {
			subsnakesToSnakeCoefMap_ = new int[totalM_];
			coefToTrack_ = new boolean[totalM_];
			coefToHide_ = new boolean[totalM_];
			coefToFreeze_ = new boolean[totalM_];
		}
		
		ArrayList<Snake2DNode> tempCoef = new ArrayList<Snake2DNode>();
		
		int backupM = M_;
		
		int mapIdx = 0;
		for (int i = 0 ; i < nSubsnakes_ ; i++) {
			
			Snake2DNode[] currentSubsnakeNodes = subsnakes_[i].getNodes();
			
			for (int j = 0 ; j < allM_[i] ; j++) {
				
				Snake2DNode currentNode = currentSubsnakeNodes[j];
				
				Snake2DNode mirrorNodeOfCurrentNode = null;
				if (withIntermedSubdiv) {
					mirrorNodeOfCurrentNode = 
							getInitialMirrorNode(currentSubsnakeNodes[j], i);
				}
				
				/* if initialization step */
				if (initNodes) {
					int positionOfCurrentNode = tempCoef.indexOf(currentNode);
					boolean isNewNode = (positionOfCurrentNode == -1);
					
					int positionOfMirrorNode;
					boolean isInCommonEdge;
					if (withIntermedSubdiv && !showCommonEdges_) {
						// get the index of the mirror node
						positionOfMirrorNode = tempCoef.indexOf(mirrorNodeOfCurrentNode);
						
						/* if already present in tempCoef, this means that the current node
						 * is part of a common edge, and should (depending on controlNodeHandling_)
						 * be replaced by its 'mirror' node
						 */
						isInCommonEdge = (positionOfMirrorNode > -1);
					}
					else {
						// either common edges are not treated, either the node is not in a common edge
						isInCommonEdge = false;
						positionOfMirrorNode = -1; // won't be used anyway in this case
					}
					
					/* case 1 : node not already visited and not in a common edge */
					if (isNewNode && !isInCommonEdge) {
				
						int positionOfNewNode = tempCoef.size();
						subsnakesToSnakeCoefMap_[mapIdx] = positionOfNewNode;
						coefToTrack_[mapIdx] = true;
						coefToHide_[mapIdx] = false;
						coefToFreeze_[mapIdx] = false;
						
						// add this node to the whole snake
						tempCoef.add(currentNode);
					}
					
					/* case 2 : node not already seed BUT in a common edge */
					else if (isNewNode && isInCommonEdge) {
						
						// two options to treat this kind of control node that belongs to a common edge :
						// - simply freeze it where it stands initially (never move during optimization)
						// - always fix its position to the 'mirrored' node, i.e. its equivalent node in the kept edge
						switch(controlNodeHandling_) {
						case FREEZE:
							int positionOfNewNode = tempCoef.size();
							subsnakesToSnakeCoefMap_[mapIdx] = positionOfNewNode;
							coefToTrack_[mapIdx] = true;
							
							// freeze the node
							currentNode.freeze();
							coefToFreeze_[mapIdx] = true;
							coefToHide_[mapIdx] = false;
							
							// add this node to the whole snake
							tempCoef.add(currentNode);
							break;
						
						case HIDE_AND_FREEZE:
							positionOfNewNode = tempCoef.size();
							subsnakesToSnakeCoefMap_[mapIdx] = positionOfNewNode;
							coefToTrack_[mapIdx] = true;
							
							// freeze the node
							coefToFreeze_[mapIdx] = true;
							currentNode.freeze();
							
							// hide that control node so it does not appear in the ROI
							coefToHide_[mapIdx] = true;
							currentNode.hide();
							
							// add this node to the whole snake
							tempCoef.add(currentNode);
							break;
							
						case FIX_TO_CORRECT_EDGE:
							// map to the 'mirrored' node
							subsnakesToSnakeCoefMap_[mapIdx] = positionOfMirrorNode;
							coefToTrack_[mapIdx] = false;
							
							coefToHide_[mapIdx] = false;
							coefToFreeze_[mapIdx] = false;
							break;
							
						case NONE:
							// do exactly the same than a node out of a common edge (= case 1 above)
							positionOfNewNode = tempCoef.size();
							subsnakesToSnakeCoefMap_[mapIdx] = positionOfNewNode;
							coefToTrack_[mapIdx] = true;
							coefToHide_[mapIdx] = false;
							coefToFreeze_[mapIdx] = false;
							
							// add this node to the whole snake
							tempCoef.add(currentNode);
						}
						
					}
					
					/* case 3 : node already inside tempCoef, so create the mapping between the two nodes */
					else {
						
						subsnakesToSnakeCoefMap_[mapIdx] = positionOfCurrentNode;
						coefToTrack_[mapIdx] = false;
						coefToHide_[mapIdx] = false;
						coefToFreeze_[mapIdx] = false;
					}
				}
				
				/* if not initialization step */
				else {
					
					if (coefToTrack_[mapIdx]) {
						
						if (coefToHide_[mapIdx]) {	
							currentNode.hide();
						}
						if (coefToFreeze_[mapIdx]) {	
							currentNode.freeze();
						}
						
						tempCoef.add(currentNode);
					}
				}
				
				mapIdx++;
			}
		}
		
		M_ = tempCoef.size();
		if ((!initNodes) && (M_ != backupM)) {
			System.err.println("GlobalSnake.M_ has changed !");
			return;
		}
		coef_ = new Snake2DNode[M_];
		tempCoef.toArray(coef_); // update corresponding parent class attribute

		coefUpToDate_ = true;
	}


	/** Alias for better readability */
	private void buildControlNodes(boolean initNodes) {
		 buildControlNodesMain(initNodes, false);
	}
	
	
	/** Alias for better readability */
	private void rebuildControlNodes(boolean initNodes) {
		 buildControlNodesMain(initNodes, true);
	}
	
	
	/**
	 * Adapted version of the public version of super.initialize() (same end though)
	 */
	private void initialize() {		
		/* initialize independent subsnakes */
		createSubsnakes();
		
		/* compute coef_ */
		buildControlNodes(true);
		
		/* create the graph to eliminate common edges */
		graph_ = new GlobalSnakeGraph(this);
		/* NOTE : depending on possible future implementations,
		 * it may be better to pass a (deep) copy of 'this' to the graph's constructor,
		 * in order to prevent unwanted attributes updates
		 */
		
		/* IMPORTANT : all Subsnakes are assumed to have the same Level_ (= this.Level_) */
		boolean doIntermedSubdiv = (Level_ > intermediateSubdiv_) 
				&& (intermediateSubdiv_ > 0);
		if (doIntermedSubdiv) {
			recreateSubsnakes();
			coefUpToDate_ = false;
			rebuildControlNodes(true);
		}
		
		/* compute subdivCoef_ */
		buildSubdivisionNodes(true);
		
		/* Once subsnakes are created, tell the graph to
		 * compute the mapping between subdivision points of original subsnakes 
		 * and 'corrected' ones from the global snake architecture (common edges)
		 */
		graph_.buildSubsnakeSubdivisionMapping();
		
		/* Send a version of that global snake to each Subsnake */
		for (SubsnakeSD subsnake : subsnakes_) {
			subsnake.setGlobalSnake(this);
		}
		
		switch (BASIS_FUNCTION) {
		case DESLAURIERSDUBUC:
			NR_ = BSplineBasis.QUADRATICBSPLINESUPPORT * nSamplesPerSegment_;
			break;
		default:
			System.err.println("Basis function type not supported in GlobalSnake.");
			return;
		}
		MR_ = M_ * nSamplesPerSegment_;
		PIM_ = Math.PI / M_;
		PI2M_ = 2 * PIM_;

		xSnakeContour_ = new double[MR_];
		ySnakeContour_ = new double[MR_];
		xSnakeTangentVector_ = new double[MR_];
		ySnakeTangentVector_ = new double[MR_];

		xSnakeContourSubdivision_ = new double[MSub_];
		ySnakeContourSubdivision_ = new double[MSub_];
		xSnakeTangentVectorSubdivision_ = new double[MSub_];
		ySnakeTangentVectorSubdivision_ = new double[MSub_];

		qLUT_ = new double[M_][M_];

		snakeContourBoundingBox_ = new Rectangle();
		snakeShellBoundingBox_ = new Rectangle();
		
		life_ = maxLife_;
		buildLUTs();
		
		updateSnakeSkin(); // calls subdivision...() and computeDerivativeSubdivPoints...()
		updateAreaSubdivision();
	}

	
	/**
	 * called by createSubsnakes() & recreateSubsnakes()
	 */
	private void createSubsnakesMain(boolean withIntermedSubdiv) {
		allM_ = new int[nSubsnakes_];
		allMSub_ = new int[nSubsnakes_];
		subsnakes_ = new SubsnakeSD[nSubsnakes_];
		oldToNewCoefsMap_ = new ArrayList<int[]>();
		
		int sumM = 0;
		int sumMSub = 0;
		for (int i = 0 ; i < nSubsnakes_ ; i++) {
			
			ESnakeParametersSD params = allSubsnakeParameters_[i];
			Snake2DNode[] nodes = allSubsnakeCoefs_.get(i).toCoef();
			
			if (withIntermedSubdiv) {
				nodes  = applyIntermediateSubdivisionToNodes(nodes, params, intermediateSubdiv_);
				params = applyIntermediateSubdivisionToParams(nodes, params, intermediateSubdiv_);
			}
			else buildOldToNewMapping(nodes, nodes);
			
			/* Subsnakes creation */
			// [imageLUTs_ is already set with call to constructor super()]
			SubsnakeSD subsnake = new SubsnakeSD(imageLUTs_, params, i);
			subsnake.setNodes(nodes);
			
			subsnakes_[i] = subsnake;
			
			allM_[i] = subsnake.getNumNodes();
			sumM += allM_[i];
			
			allMSub_[i] = subsnake.getNumSubdivisionPoints();
			sumMSub += allMSub_[i];
		}
		totalM_ = sumM;
		totalMSub_ = sumMSub;
		
		if (withIntermedSubdiv) {
			// new effective subdiv level (globalLevel_ stays the same, the original global snake level)
			setLevelSubdivision(Level_ - intermediateSubdiv_);
		}
	}
	
	
	private void createSubsnakes() {
		createSubsnakesMain(false);
	}
	
	
	private void recreateSubsnakes() {
		createSubsnakesMain(true);
	}


	private Snake2DNode[] applyIntermediateSubdivisionToNodes(Snake2DNode[] oldNodes, 
			ESnakeParametersSD oldParams, int numInterSubdiv) {
		
		int oldM = oldNodes.length;
		ESnakeParametersSD tempParam = new ESnakeParametersSD(
				oldParams.getDetectType(),
				oldM, 
				numInterSubdiv, 
				oldParams.getBasisFunctionType(), 
				oldParams.getMScalingFactor(),
				oldParams.getEnergyType(),
				oldParams.getAlpha(),
				oldParams.getPriorShapeType(), 
				oldParams.getMaxLife(),
				oldParams.isImmortal(), 
				oldParams.getMultiresolutionType());
		
		ESnakeSD tempSnake = new ESnakeSD(imageLUTs_, tempParam, null, null);
		tempSnake.setNodes(oldNodes);
		
		// subdivision nodes of that snake will become control nodes of the subsnake
		Snake2DNode[] newNodes = tempSnake.getSubdivisionPoints();
		
		// build mapping of *old* vs *new* control nodes
		buildOldToNewMapping(oldNodes, newNodes);
		
		return newNodes;
	}
	
	
	private ESnakeParametersSD applyIntermediateSubdivisionToParams(Snake2DNode[] newNodes, 
			ESnakeParametersSD oldParams, int numInterSubdiv) {
		
		// new effective subdiv level and number of control nodes
		int adaptedLevel = oldParams.getSubdivisionLevel() - numInterSubdiv;
		int adaptedM = newNodes.length;
		
		ESnakeParametersSD newParams = new ESnakeParametersSD(
				oldParams.getDetectType(),
				adaptedM, 
				adaptedLevel, 
				oldParams.getBasisFunctionType(), 
				oldParams.getMScalingFactor(),
				oldParams.getEnergyType(),
				oldParams.getAlpha(),
				oldParams.getPriorShapeType(), 
				oldParams.getMaxLife(),
				oldParams.isImmortal(), 
				oldParams.getMultiresolutionType());
		
		return newParams;
	}
	
	
	/**
	 * Compute the mapping between original control nodes and last version of it
	 * (if no intermed subdiv, the mapping is trivially a 1-to-1 ordered relation)
	 */
	private void buildOldToNewMapping(Snake2DNode[] oldNodes, Snake2DNode[] newNodes) {
		int oldM = oldNodes.length;
		int[] map = new int[oldM];
		for (int j = 0 ; j < oldM ; j++) {
			int index = Arrays.asList(newNodes).indexOf(oldNodes[j]);
			if (index > -1) map[j] = index;
			else System.err.println("Warning : unable to map original to "
					+ "new nodes after intermed subdiv !");
		}
		oldToNewCoefsMap_.add(map);
	}
	
	
	/** TODO : add doc */
	private boolean belongToCleanEdge(Snake2DNode node) {
		return graph_.belongToCleanEdge(node);
	}
	
	
	/** TODO : add doc */
	private Snake2DNode getInitialMirrorNode(Snake2DNode inputNode, int subsnakeId) {
		return (Snake2DNode) graph_.getMirrorNode(inputNode, subsnakeId);
	}
	
	
	/**
	 * Update the fields of allSubsnakeCoefs to match the current configuration 
	 * of individual subsnakes, in order to start from that new state when 
	 * changing parameters from the settings panel
	 */
	private void refreshInitCoefs() {
		for (int i = 0 ; i < nSubsnakes_ ; i++) {
			final int origM = allSubsnakeCoefs_.get(i).size();
			final Point2D.Double seedPoint = 
					allSubsnakeCoefs_.get(i).getStartingPoint();
			
			Snake2DNode[] origNodes = new Snake2DNode[origM];
			final Snake2DNode[] currNodes = subsnakes_[i].getNodes();
			
			int[] map = oldToNewCoefsMap_.get(i);
			for (int j = 0 ; j < origM ; j++) {
				origNodes[j] = currNodes[map[j]];
			}
			allSubsnakeCoefs_.set(i, new SubsnakeInitialShape(origNodes, seedPoint));
		}
	}
	
	
	private double sumEnergy() {
		double sumE = 0;
		for (int i = 0 ; i < nSubsnakes_ ; i++) {
			sumE += subsnakes_[i].energy();
		}
		if (!immortal_) {
			life_--;
			if (life_ == 0) {
				alive_ = false;
			}
		}
		return sumE;
	}
	
  
	// ----------------------------------------------------------------------------
	// Convenience functions
 
//	/**
//	 * Convenience functions to convert T[] into ArrayList<T>.
//	 */
//	private static <T> ArrayList<T> array2ArrayList(T[] array) {
//		ArrayList<T> list = new ArrayList<T>();
//		for (T i : array) list.add(i);
//		return list;
//	}
//  
	
//	/**
//	 * For debugging purposes, print index of control nodes
//	 * and there equivalent in the subdivision nodes space.
//	 */
//	private void debugPrintNodesIndices(int subsnakeId) {
//		if (subsnakeId >= nSubsnakes_) return;
//    
//		int i = subsnakeId;
//		Snake2DNode[] ssnakeCoefs = subsnakes_[i].getNodes();
//		for (int n = 0 ; n < subsnakes_[i].getNumNodes() ; n++) {
//			System.out.print(n + " : ");
//			ArrayList<Snake2DNode> subdivCoefList = array2ArrayList(subsnakes_[i].getSubdivisionPoints());
//		}
//	}
}
