package plugins.big.bigsnake.snake;

import java.awt.geom.Point2D;
import java.awt.geom.Point2D.Double;
import java.util.ArrayList;
import java.util.Comparator;

import plugins.big.bigsnakeutils.icy.snake2D.Snake2DNode;
import plugins.big.bigsnakeutils.process.process1D.BSplineBasis;

/**
 * Class extending a simple list of nodes, adding the centroid of the
 * initial node configuration, in order to make use of a Comparator
 * to order a list if <code>SubsnakeInitialShape</code> 
 * by the x/y positions of centroids.
 * 
 * NOTE : it does not extend the PriorShape class from Active Cells SDK,
 * because its goal is not really the same (hence a different name).
 * 
 * 
 * @version January 13, 2018
 * 
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 *
 */
public class SubsnakeInitialShape extends ArrayList<Snake2DNode> {
	
	private static final long serialVersionUID = 83864890754805572L;
	
	/** Centroid of the Subsnake (barycenter) */
	private Point2D.Double centroid_ = null;

	/** Area of the Subsnake */
	private double area_ = 0;
	
	/** Optional, point that were used to create the Subsnake shape */
	private Point2D.Double startingPoint_ = null;
	
	public SubsnakeInitialShape() {
		super();
	}
	
	public SubsnakeInitialShape(Point2D.Double startingPoint) {
		super();
		startingPoint_ = startingPoint;
	}
	
	public SubsnakeInitialShape(Snake2DNode[] nodeArray, Point2D.Double startingPoint) {
		clear();
		for (Snake2DNode node : nodeArray) {
			super.add(node);
		}
		startingPoint_ = startingPoint;
		updateAllFields();
	}
	
	public SubsnakeInitialShape deepCopy(SubsnakeInitialShape shape) {
		SubsnakeInitialShape shapeCopy = new SubsnakeInitialShape(shape.getStartingPoint());
		for (Snake2DNode node : shape) {
			shapeCopy.add(new Snake2DNode(node));
		}
		return shapeCopy;
	}

	public double getStartingPointX() {
		return startingPoint_.x;
	}
	
	public double getStartingPointY() {
		return startingPoint_.y;
	}
	
	public double getCentroidX() {
		return centroid_.x;
	}
	
	public double getCentroidY() {
		return centroid_.y;
	}
	
	public double getArea() {
		return area_;
	}
	
	public Point2D.Double getStartingPoint() {
		return startingPoint_;
	}
	
	public Snake2DNode[] toCoef() {
		Snake2DNode[] coef = new Snake2DNode[size()];
		return this.toArray(coef);
	}
	
	@Override
	public boolean add(Snake2DNode node) {
		boolean result = super.add(node);
		updateAllFields();
		return result;
	}
	
	/**
	 * Special sorting method that puts the node corresponding 
	 * the comparator criterion at the beginning of the list,
	 * while keeping the order of the nodes, to ensure a closed shape
	 * (basically performs a cyclic permutation)
	 */
	public void sortAndPreserveShape(Comparator<Snake2DNode> c) {
		SubsnakeInitialShape original = deepCopy(this);
		
		SubsnakeInitialShape temp = deepCopy(this);
		temp.sort(c);
		
		/* The first element of temp is now the one to keep 
		 * at that position. We now need to preserve original order.
		 */
		
		// get the index of the first element in the original shape
		final int idxFirstNode = original.indexOf(temp.get(0));
		
		/* finally set the new nodes at their new position */
		int idx = idxFirstNode;
		for (int i = 0 ; i < size() ; i++) {
			set(i, original.get(idx % size()));
			idx++;
		}
	}
	
	private void updateAllFields() {
		updateCentroid();
		updateArea();
	}
	
	/** Adapted from  {@link plugins.big.bigsnake.snake.ESnakeSD#getCentroid()} */
	private void updateCentroid() {
		Point2D.Double centroid = new Point2D.Double();
		for (Snake2DNode element : this) {
			centroid.x += element.x;
			centroid.y += element.y;
		}
		centroid.x /= size();
		centroid.y /= size();
		centroid_ = centroid;
	}
	
	/** Adapted from {@link plugins.big.bigsnake.snake.ESnakeSD#updateArea()} */
	private void updateArea() {
		int numNodes = size();
		
		int qSize = 2 * BSplineBasis.ESPLINE3SUPPORT - 1;
		double[] bSplineAutocorrelationLUT = new double[qSize];
		for (int i = 0; i < qSize; i++) {
			bSplineAutocorrelationLUT[i] = BSplineBasis.correlationESpline(i
					- BSplineBasis.ESPLINE3SUPPORT + 1, 2*Math.PI/numNodes);
		}
		
		double area = 0;
		for (int k = 0; k < numNodes; k++) {
			int kN = k + BSplineBasis.ESPLINE3SUPPORT;
			for (int l = k - BSplineBasis.ESPLINE3SUPPORT + 1; l < kN; l++) {
				int l_p = l;
				while (l_p < 0) {
					l_p += numNodes;
				}
				while (l_p >= numNodes) {
					l_p -= numNodes;
				}
				area += get(k).y * get(l_p).x
						* bSplineAutocorrelationLUT[kN - l - 1];
			}
		}
		area_ = area;
	}

}
