package plugins.big.bigsnake.snake;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.w3c.dom.Node;

import plugins.big.bigsnake.core.ImageLUTContainerSD;
import plugins.big.bigsnake.graph.GlobalSnakeGraph;
import plugins.big.bigsnake.snake.Enums.ESnakeEnergyTypeSD;
import plugins.big.bigsnake.snake.Enums.ESnakeTargetTypeSD;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DNode;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DScale;
import plugins.big.bigsnakeutils.process.process1D.BSplineBasis;
import plugins.big.bigsnakeutils.process.process1D.BSplineBasis.BSplineBasisType;
import plugins.big.bigsnakeutils.shape.priorshapes.ShapeProjectorBuilder;
import plugins.big.bigsnakeutils.shape.priorshapes.shapes.Custom;

/**
 * Class that describes a Subsnake (= component of a GlobalSnake)
 * (based on Exponential spline snake).
 * 
 * @version January 13, 2018
 * 
 * @author Ricard Delgado-Gonzalo (ricard.delgado@gmail.com)
 * @author Ramtin Madani (ramtin_madani@yahoo.com)
 * @author Masih Nilchian (masih_n85@yahoo.com)
 * @author Daniel Schmitter (daniel.schmitter@epfl.ch)
 * @author Ana�s Badoual (anais.badoual@gmail.com)
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 */
public class SubsnakeSD extends ESnakeSD {
	
	/** 
	 * Unique ID related to the sub cell in the image.
	 */
	private int subsnakeId_ = 0;
	
	/**
	 * Simply initial control and subdivision nodes as List
	 * (mainly for practicality)
	 */
	private List<Snake2DNode> initCoefAsList_ = null;
	private List<Snake2DNode> initSubdivCoefAsList_ = null;
	// flag that tells if the nodes were changed or not
	private boolean firstSetNodes_ = true;
	
	/**
	 * Samples of the x & y coordinates of the global snake contour obtained by
	 * subdivision.
	 */
	private double[] xSubsnakeContourSubdivision_ = null;
	private double[] ySubsnakeContourSubdivision_ = null;

	/**
	 * Stores the hidden state of subdivision Snake2DNodes
	 * to re-establish them when nodes are re-computed
	 */
	private boolean[] subdivHiddenStates_ = null;
	
	/**
	 * GlobalSnake object from where this Subsnake was created.
	 */
	private GlobalSnakeSD parentGlobalSnake_ = null;
	
	
	// ============================================================================
	// PUBLIC METHODS
	
	
	public SubsnakeSD(ImageLUTContainerSD imageLUTs, ESnakeParametersSD parameters, Object initialContour,
			Custom customPriorShape, int subsnakeId) {
		
		super(imageLUTs, parameters, initialContour, customPriorShape);
		
		/* NOTE : after a super() call, instance fields above are instantiated again
		 * (weird, but that's how Java rolls)
		 */
		
		SNAKE_CONTOUR_COLOR = new Color(255, 0, 0, 255);
		CONTROL_POLYGON_COLOR = new Color(255, 255, 0, 100);  // dim a little bit the polygon yellow line
		
		subsnakeId_ = subsnakeId;
		
		xSubsnakeContourSubdivision_ = new double[MSub_];
		ySubsnakeContourSubdivision_ = new double[MSub_];
		
		convertCoefToList();
	}
	
	
	public SubsnakeSD(ImageLUTContainerSD imageLUTs, ESnakeParametersSD parameters, int subsnakeId) {
		this(imageLUTs, parameters, null, null, subsnakeId);
	}
	
	
	// TODO : change return type to Graph.NodeList ?
	public List<Snake2DNode> getInitNodes() {
		return initCoefAsList_;
	}
	
	
	public List<Snake2DNode> getInitSubdivisionPoints() {
		return initSubdivCoefAsList_;
	}
	
	
	public int getNumSubdivisionPoints() {
		return subdivCoef_.length;
	}
	
	
	public int getId() {
		return subsnakeId_;
	}
	
	
	public void setGlobalSnake(GlobalSnakeSD globalSnake) {
		parentGlobalSnake_ = globalSnake;
	}
	
	
	@Override
	public void setNodes(Snake2DNode[] node) {
		super.setNodes(node);
		
		if (subdivHiddenStates_ != null) {
			for (int i = 0 ; i < MSub_ ; i++) {
				if (subdivHiddenStates_[i] == true) {
					subdivCoef_[i].hide();
				}
				else {
					subdivCoef_[i].unhide();
				}
			}
		}
		
		if (firstSetNodes_) {
			convertCoefToList();
			firstSetNodes_ = false;
		}
	}
	
	
	public void setEnergyType(ESnakeEnergyTypeSD energyType) {
		energyType_ = energyType;
	}
	
	
	public void setMaxLife(int maxLife) {
		maxLife_ = maxLife;
	}

	
	public void setBasisFunctionType(BSplineBasisType basisFunctionType) {
		basisFunctionType_ = basisFunctionType;
		BASIS_FUNCTION = basisFunctionType;
	}

	
	public void setImmortal(boolean immortal) {
		immortal_ = immortal;
	}

	
	public void setDetectType(ESnakeTargetTypeSD detectType) {
		detectType_ = detectType;
	}
	
	
	@Override
	public Snake2DScale[] getScales() {
		
		Snake2DScale[] skin = super.getScales();
		
		// replace the default red smooth curve of the skin by a custom one
		skin[0] = new Snake2DScale(xSubsnakeContourSubdivision_, ySubsnakeContourSubdivision_, 
				xSubsnakeContourSubdivision_.length, SNAKE_CONTOUR_COLOR, true);
		
		return skin;
	}
	
	
	@Override
	public Snake2DNode[] getSmoothSkin() {
		Snake2DNode[] smoothSkin = new Snake2DNode[MSub_];
		for (int i = 0 ; i < MSub_ ; i++) {
			smoothSkin[i] = new Snake2DNode(
					xSubsnakeContourSubdivision_[i], 
					ySubsnakeContourSubdivision_[i]);
		}
		return smoothSkin;
	}
	
	
	public void saveHiddenStates() {
		subdivHiddenStates_ = new boolean[MSub_];
		for (int i = 0 ; i < MSub_ ; i++) {
			subdivHiddenStates_[i] = subdivCoef_[i].isHidden();
		}
	}
	

	// ============================================================================
	// PROTECTED METHODS
	
	@Override
	protected void initializeContoursAndTangents() {
		super.initializeContoursAndTangents();
		xSubsnakeContourSubdivision_ = new double[MSub_];
		ySubsnakeContourSubdivision_ = new double[MSub_];
	}
	
	
	@Override
	protected void updateSnakeSkin() {
		super.updateSnakeSkin();
		updateSubsnakeSkin();
	}
	
	
	// ============================================================================
	// PRIVATE METHODS
	
	
	private void updateSubsnakeSkin() {
		
		for (int i = 0 ; i < MSub_ ; i++) {
			
			if (parentGlobalSnake_ != null) {
				
				Snake2DNode correctNode = parentGlobalSnake_
						.getCorrectSubdivisionNode(subsnakeId_, i);
				
				if (correctNode == null) {
					// if the methods above returns null, this means 
					// that the node is in fact the the subdivCoef, as usual
					correctNode = subdivCoef_[i];
				}
				xSubsnakeContourSubdivision_[i] = correctNode.x;
				ySubsnakeContourSubdivision_[i] = correctNode.y;
			}
			
			else {
				xSubsnakeContourSubdivision_[i] = xSnakeContourSubdivision_[i];
				ySubsnakeContourSubdivision_[i] = ySnakeContourSubdivision_[i];
			}
		}
	}
	
	
	// ================
	
	
	private void convertCoefToList() {
		initCoefAsList_ = Arrays.asList(coef_);
		initSubdivCoefAsList_ = Arrays.asList(subdivCoef_);
	}

}
