package plugins.big.bigsnake.tests;

import java.io.IOException;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Map;

import plugins.big.bigsnake.core.ImageLUTContainerSD;
import plugins.big.bigsnake.jaccard.JaccardSubsnakeStatistics;
import plugins.big.bigsnake.nuclei.NucleiDetector;
import plugins.big.bigsnake.snake.GlobalSnakeInitializer;
import plugins.big.bigsnake.snake.GlobalSnakeParametersSD;
import plugins.big.bigsnake.snake.GlobalSnakeSD;
import plugins.big.bigsnake.snake.SubsnakeSD;
import plugins.big.bigsnake.utils.FileUtils;
import plugins.big.bigsnake.utils.ImageUtils;
import plugins.big.bigsnake.voronoi.VoronoiDiagram;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DNode;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DOptimizer;
import plugins.big.bigsnakeutils.icy.snake2D.Snake2DPowellOptimizer;

public class GlobalSnakeTests {
	
	private final static int MIN_INDEX = 0;
	private final static int MAX_INDEX = 1000;  // inclusive, huge value = process all files
	private static final boolean SUB_OPTIMIZE = true;
	private static final boolean SUBSUB_OPTIMIZE = false;

	private final static String WORKING_DIR = 
			"/Users/anais/Documents/workspaceAymeric/ssnakebig/test_images/ideal_emb1_z36_t22";

	private final static String MASK_FOLDER = "masks";
	private final static String IMAGE_FOLDER = "basetest_gaussnoise";
	private final static String SEED_FOLDER = "nuclei";
	private final static String OUTPUT_FOLDER = "results";
	
	private final static String[] EXTENSIONS = {"jpg", "png", "tif"};
	
	private static TestParameters param_ = null;
	private static GlobalSnakeInitializer initializer_ = null;
	private static GlobalSnakeSD snake_ = null;
	private static String imageDirectory_ = "/";
	private static String maskDirectory_ = "/";
	
	private static TestOutputWriter testWriter_ = null;
	
	/* internal use */
	private static int subRunCounter_ = 0;
	
	
	public static void main(String[] args) throws Exception {
		
		final double startTime = System.currentTimeMillis();
		
		initialize();
		
		run();
		
		finish();
		
		System.out.println("\nPipeline complete.");
		
		final double endTime = System.currentTimeMillis();
		
		final double totalTimeSec = (endTime - startTime) / 1000.;
		final int seconds = ((int) totalTimeSec) % 60;
		final int minutes =  ((int) (totalTimeSec) % 3600) / 60;
		final int hours = ((int) totalTimeSec) / 3600;
		
		System.out.println(String.format("Total pipeline duration : %d:%02d:%02d", hours, minutes, seconds));
	}
	

	private static void run() throws Exception {
		/* Retrieve all image files */
		final String[] imageFilenames = 
				FileUtils.getAllFilenamesInDirectory(imageDirectory_, EXTENSIONS);
		
		/* Ensure files are sorted */
		Arrays.sort(imageFilenames);
		
		/* Number of iterations in the loop over images */
		final int minIndex = Math.max(MIN_INDEX, 0);
		final int maxIndex = Math.min(MAX_INDEX, imageFilenames.length-1);
		System.out.println("Number of iterations to run : " + (maxIndex-minIndex+1));
		
		/* ======== MAIN LOOP ======== */
		int iterNum = 1;
		for (int iterIdx = minIndex ; iterIdx <= maxIndex ; iterIdx++) {
			
			System.out.println("\n========== start iteration " + iterNum + " ==========");
			
			if (iterNum == 1) {
				/* Compute Jaccard indices with NO OPTIMIZATION (only initial shape) */
				initializeOptimization();
				blankRun(imageFilenames[0]);
				computeJaccardIndex("initialization_only", 0, true);
			}
			
			final String filename = imageFilenames[iterIdx];
			System.out.println("Embryo image : " + filename);
			
			/* Nuclei detection + control nodes initialization */
			initializeOptimization();
			
			/* Optimization(s) */
			
			subRunCounter_ = 0; // reset counter
			
			final double optimizationStartTime = System.currentTimeMillis();

			subRun(filename);
			
			if (SUB_OPTIMIZE) {
				
				subRun(filename);
				
				if (SUBSUB_OPTIMIZE) {
					
					subRun(filename);
					
				}
			}
			
			final double optimizationTimeSec = 
					(System.currentTimeMillis() - optimizationStartTime) / 1000.;
			
			computeJaccardIndex(filename, optimizationTimeSec, true);
			
			System.out.println("=========== end iteration " + iterNum + " ===========");
			iterNum++;
			
//			Thread.sleep(2000); // milliseconds
		}
		/* ======== end main loop ======== */
		
	}
	
	
	/**
	 * Used to compute Jaccard index of the initial configuration 
	 * (no optimization performed)
	 */
	private static void blankRun(String filename) throws IOException {
		
		final String path0 = Paths.get(imageDirectory_, filename).toString();
		
		createNewGlobalSnake(path0);
	}
	
	
	/**
	 * Initialization and optimization of a GlobalSnake from the parameters (param_), 
	 * performed on the specified image (filename).
	 * Depending on the subRunCounter, it runs the first optimization, 
	 * a second (sub-) optimization from finer parameters,
	 * and a third (sub-sub-) one from even finder parameters.
	 */
	private static void subRun(String filename) throws IOException {
		
		final String path = Paths.get(imageDirectory_, filename).toString();
		
		if (subRunCounter_ == 0) {
			
			createNewGlobalSnake(path);
			
			final double startTime = System.currentTimeMillis();
			optimizeSnake();
			final double endTime = System.currentTimeMillis();
			System.out.println("---> Optimization duration : " + 
					((endTime - startTime) / 1000.) + " s");
		}
		
		else if (subRunCounter_ == 1) {
			
			updateCurrentGlobalSnake(path);
			
			final double startTime = System.currentTimeMillis();
			optimizeSnake();
			final double endTime = System.currentTimeMillis();
			System.out.println("---> Sub-optimization duration : " + 
					((endTime - startTime) / 1000.) + " s");
		} 
		
		else {
			
			updateCurrentGlobalSnake(path);
			
			final double startTime = System.currentTimeMillis();
			optimizeSnake();
			final double endTime = System.currentTimeMillis();
			System.out.println("---> Sub-sub-optimization duration : " + 
					((endTime - startTime) / 1000.) + " s");
		}
		
		subRunCounter_++;
	}


	private static void computeJaccardIndex(String filename,
			double runtimeSeconds, boolean csvOutput) throws IOException {
		List<SubsnakeSD> subsnakeList = new ArrayList<SubsnakeSD>();
		for (int id = 0 ; id < snake_.getNumSubsnakes() ; id++) {
			subsnakeList.add(snake_.getSubsnake(id));
		}
		
		JaccardSubsnakeStatistics jaccard = new JaccardSubsnakeStatistics(subsnakeList, 
				initializer_.getOrderedSeeds());
		jaccard.setMaskDirectory(maskDirectory_);
		jaccard.run(false, param_.jaccardMode);
		
		if (csvOutput) {
			final Map<String, Double> all = jaccard.getAllIndices();
			final double average = jaccard.getAverage();
			final double std = jaccard.getStandardDeviation();
			testWriter_.outputJaccardStatistics(filename, average, std, all, runtimeSeconds);
		}
	}


	private static void createNewGlobalSnake(String imagePath) {
		ImageLUTContainerSD imageContainter = new ImageLUTContainerSD();
		try {
			imageContainter.setSequence(ImageUtils.getSequenceLikeInIcy(imagePath));
		} catch (IOException e1) {
			System.err.println("Couldn't load test image sequence");
			e1.printStackTrace();
			return;
		}
		imageContainter.setChannelNumber(param_.imageChannel);
		imageContainter.setSigma(param_.imageSigma);
		try {
			imageContainter.buildLUTs(false); // false for no GUI
		} catch (Exception e) {
			e.printStackTrace();
			return;
		}
		System.out.println("LUTs smoothing : " + imageContainter.getSigma());
		
		final GlobalSnakeParametersSD snakeParameters = new GlobalSnakeParametersSD(
				initializer_.getInitialShapes(), param_.subdivLevel, 
				param_.targetBrightness, param_.basisFunction, param_.energy, 
				param_.globalEnergy, param_.maxLife, param_.immortal, 
				param_.commonEdges, param_.intermedSubdiv);
		
		snake_ = new GlobalSnakeSD(imageContainter, snakeParameters);
		System.out.println(snake_);
	}
	
	
	private static void updateCurrentGlobalSnake(String imagePath) {
		
		ImageLUTContainerSD imageContainter = snake_.getImageLUTs();
		
		// update few parameters
		if (subRunCounter_ == 1) {
			imageContainter.setSigma(param_.imageSigmaSub);
		} else if (subRunCounter_ == 2) {
			imageContainter.setSigma(param_.imageSigmaSubSub);
		}
		try {
			imageContainter.buildLUTs(false); // false for no GUI
		} catch (Exception e) {
			e.printStackTrace();
			return;
		}
		System.out.println("LUTs smoothing : " + imageContainter.getSigma());
		
		GlobalSnakeParametersSD snakeParameters = 
				(GlobalSnakeParametersSD) snake_.getSnakeParameters();
		
		// update few parameters
		if (subRunCounter_ == 1) {
			snakeParameters.setIntermediateSubdiv(param_.intermedSubdivSub);
			snakeParameters.setMaxLife(param_.maxLifeSub);
			snakeParameters.setImmortal(param_.immortalSub);
		} else if (subRunCounter_ == 2) {
			snakeParameters.setIntermediateSubdiv(param_.intermedSubdivSubSub);
			snakeParameters.setMaxLife(param_.maxLifeSubSub);
			snakeParameters.setImmortal(param_.immortalSubSub);
		}
		
		snake_ = new GlobalSnakeSD(imageContainter, snakeParameters);
		System.out.println(snake_);
	}


	private static void initialize() {
		/* Load parameters */
		param_ = new TestParameters();
		
		/* Build directories */
		imageDirectory_ = Paths.get(WORKING_DIR, IMAGE_FOLDER).toString();
		maskDirectory_ = Paths.get(WORKING_DIR, MASK_FOLDER).toString();
		
		/* Create CSV output file */
		final String outputFilename = String.format("jacc_%s_%s.csv", IMAGE_FOLDER, getDateTime());
		final String csvPath = Paths.get(WORKING_DIR, OUTPUT_FOLDER, outputFilename).toString();
		testWriter_ = new TestOutputWriter(csvPath, true);
	}


	private static void initializeOptimization() {
		/* Nuclei detector */
		final String nucleiFilename = "nuclei.jpg";
		final String nucleiPath = Paths.get(WORKING_DIR, SEED_FOLDER, nucleiFilename).toString();
		System.out.println("Nuclei image : " + nucleiFilename);
		
		NucleiDetector nucleiDetector = new NucleiDetector(false); // false when launching out of Icy GUI
		try {
			nucleiDetector.setSequence(ImageUtils.getSequenceLikeInIcy(nucleiPath));
		} catch (IOException e) {
			System.err.println("Couldn't load nuclei sequence");
			e.printStackTrace();
			return;
		}
		nucleiDetector.setSigmaLoG(param_.detectorSigma);
		nucleiDetector.setThreshold(param_.detectorThreshold);
		nucleiDetector.run();
		
		final int numNuclei = nucleiDetector.getNucleiPositions().size();
		if (numNuclei == 2) {
			System.err.println("Less than 3 nuclei were detected");
			return;
		}
		System.out.println(numNuclei + " nuclei have been detected");
		
		/* Voronoi diagram */
		final VoronoiDiagram voronoi = new VoronoiDiagram(
				nucleiDetector.getWidth(), 
				nucleiDetector.getHeight(), 
				nucleiDetector.getNucleiPositionsAsArray(),
				param_.dilationCoef);
		
		/* GlobalSnake initialize */
		initializer_ = new GlobalSnakeInitializer(
				param_.initType, param_.sortingType, voronoi);
	}


	private static String getDateTime() {
		DateFormat dateFormat = new SimpleDateFormat("yyMMdd-HHmm");
		return dateFormat.format(new Date());
	}


	private static void finish() {
		testWriter_.finish();
	}
	
	
	private static void optimizeSnake() {
		
		Snake2DOptimizer optimizer = new Snake2DPowellOptimizer();
		
		final Snake2DNode[] youngSnake = snake_.getNodes();
		
		if (Double.isNaN(youngSnake[0].x)) {System.out.println("NaN value");}
		
		final int K = youngSnake.length;
		final Snake2DNode[] X = new Snake2DNode[K];
		for (int k = 0; k < K; k++) {
			X[k] = new Snake2DNode(youngSnake[k].x, youngSnake[k].y,
					youngSnake[k].isFrozen(), youngSnake[k].isHidden());
		}
		snake_.setNodes(X);
		snake_.reviveSnake();
		
		optimizer.optimize(snake_, X);
	}

}
