package plugins.big.bigsnake.tests.watersheds;

import java.io.IOException;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import ij.IJ;
import ij.ImagePlus;
import ij.process.ImageProcessor;
import inra.ijpb.watershed.MarkerControlledWatershedTransform2D;
import plugins.big.bigsnake.jaccard.JaccardMaskManager;
import plugins.big.bigsnake.tests.TestOutputWriter;
import plugins.big.bigsnake.utils.FileUtils;
import plugins.big.bigsnake.utils.ImageUtils;


/**
 * 
 * Automatized test pipeline to run the marker-controlled watersheds algorithm
 * by MorphoLibJ on same test images than the GlobalSnake test pipeline.
 * 
 * Attention : the Jaccard binary masks (groundtruth) are paired to regions
 * in the segmented image by using the fact that the algorithm assign 
 * colors to the regions in shade of grays consistently with the seed 
 * image provided initially.
 * You may want to change the algorithm : just make sure that segmented regions
 * have a color that is consistent with those of the seed pixels
 * (not necessarily same colors, but same ordering, from darker to lighter)
 * 
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 *
 */
public class MorphoLibWatershedTests {
	
	private static final int CONNECTIVITY = 8;
	
	/* IMPORTANT : color of the segmented region that is not part of the embryo (= background) */
	private final static double BG_COLOR = 255.0;
	
	private final static String WORKING_DIR = 
			"/Users/anais/Documents/workspaceAymeric/ssnakebig/test_images/ideal_emb1_z36_t22";
	private final static String WSHED_DIR = 
			"/Users/anais/Documents/workspaceAymeric/ssnakebig/test_images/ideal_emb1_z36_t22/Watersheds";

	private final static String MASK_FOLDER = "masks";
	private final static String IMAGE_FOLDER = "basetest_attenuation";
	private final static String SEED_FOLDER = "seeds";
	private final static String OUTPUT_FOLDER = "results";
	private final static String OUTPUT_BASENAME = "wshedtest";
	
	private final static String[] EXTENSIONS = {"jpg", "png", "tif"};
	
	private static String markerFilename_ = "watershed_markers_pixGreys.png";
	
	private static TestOutputWriter testWriter_ = null;

	private static String watershedMarkersPath_;
	private static ImageProcessor watershedMarkers_;
	private static ImageProcessor watershedMask_;
	private static ImageProcessor result_ = null;
	private static ImageProcessor[] resultCells_ = null;

	private static int width_ = 0;
	private static int height_ = 0;
	private static int numCells_ = 0;
//	private static TreeMap<Double, double[][]> colorToBinaryMaskMap_ = null;
	private static List<double[][]> jaccardMaskListSorted_ = null;
	
	
	public static void main(final String[] args) throws IOException {

		final double startTime = System.currentTimeMillis();
		
		initialize();
		
		run();
		
		finish();
		
		System.out.println("\nPipeline complete.");
		
		final double endTime = System.currentTimeMillis();
		
		final double totalTimeSec = (endTime - startTime) / 1000.;
		final int seconds = ((int) totalTimeSec) % 60;
		final int minutes =  ((int) (totalTimeSec) % 3600) / 60;
		final int hours = ((int) totalTimeSec) / 3600;
		
		System.out.println(String.format("Total pipeline duration : %d:%02d:%02d", hours, minutes, seconds));
	}
	
	
	private static void initialize() {
		/* Retrieve input image and optional mask */
		watershedMarkersPath_ = Paths.get(WSHED_DIR, SEED_FOLDER, markerFilename_).toString();
		watershedMarkers_ = IJ.openImage(watershedMarkersPath_).getProcessor();
		watershedMask_ = null; // no specific mask for proper comparison with GlobalSnake
		
		width_ = watershedMarkers_.getWidth();
		height_ = watershedMarkers_.getHeight();
		
		/* Build list of binary mask pixel arrays */
		retrieveAndSortJaccardMasks();
		
		/* Create CSV output file */
		final String outputFilename = String.format("jacc_%s_%s.csv", IMAGE_FOLDER, getDateTime());
		final String csvPath = Paths.get(WSHED_DIR, OUTPUT_FOLDER, outputFilename).toString();
		testWriter_ = new TestOutputWriter(csvPath, false);
	}
	
	
	private static void run() throws IOException {
		
		/* Get the list of test image files */
		final String inputDirectory = Paths.get(WORKING_DIR, IMAGE_FOLDER).toString();
		final String[] imageFilenames = 
				FileUtils.getAllFilenamesInDirectory(inputDirectory, EXTENSIONS);
		
		/* Ensure the ordering of files (requires good names though) */
		Arrays.sort(imageFilenames);
		
		for (String filename : imageFilenames) {
			
			System.out.println("Embryo image : " + filename);
			
			final double startTime = System.currentTimeMillis();
			
			subRun(filename);
			
			final double endTime = System.currentTimeMillis();
			final double runtimeSec = (endTime-startTime)/1000.;
		
			final String resultPath = Paths.get(WSHED_DIR, OUTPUT_FOLDER, OUTPUT_BASENAME + "_result").toString();
			
			/* Save resulting segmentation on disk */
			IJ.saveAs(new ImagePlus("result", result_), "png", resultPath);
		
			/* Separate regions into binary masks for each embryo cells */
			resultCells_ = WatershedUtils.separateCells(result_, BG_COLOR, true);
			numCells_ = resultCells_.length;
			
			/* Save binary masks on disk */
			for (int i = 0 ; i < numCells_ ; i++) {
				final String cellPath = Paths.get(WSHED_DIR, OUTPUT_FOLDER, 
						OUTPUT_BASENAME + "_c" + (i+1)).toString();
				IJ.saveAs(new ImagePlus("binary cell", resultCells_[i]), "png", cellPath);
			}
			
			/* Jaccard index computation and output */
			computeJaccardStatistics(filename, runtimeSec, true);
		}
	}


	private static void subRun(String filename) {
		final String imagePath = Paths.get(WORKING_DIR, IMAGE_FOLDER, filename).toString();
		
		ImageProcessor input = IJ.openImage(imagePath).getProcessor();
		if ((input.getWidth() != width_) || (input.getHeight() != height_)) {
			System.err.println("All input, mask, seeds, etc. images must have the same dimensions");
			return;
		}
		
		/* Initialize the marker-controlled watershed algorithm from MorphoLibJ */
		MarkerControlledWatershedTransform2D watershed = 
				new MarkerControlledWatershedTransform2D(
						input, watershedMarkers_, watershedMask_, CONNECTIVITY);
		
		/* Run the computation */
		result_ = watershed.applyWithPriorityQueue();
	}


	private static void finish() {
		testWriter_.finish();
	}


	/**
	 * 
	 */
	private static void retrieveAndSortJaccardMasks() {
		final String maskDirectory = Paths.get(WORKING_DIR, MASK_FOLDER).toString();
		final String[] maskFilenames = 
				FileUtils.getAllFilenamesInDirectory(maskDirectory, null);
		
		/* Associate each mask to the seed, so we can sort the list following seed color */
		final Map<Double, double[][]> colorToMaskMap = computeColorToMaskMapping(maskFilenames);
		
		/* Sort the mask pixel arrays according to the color value 
		 * (performed by the TreeMap initialization, with colors as keys) */
		final TreeMap<Double, double[][]> colorToMaskMapSorted = 
				new TreeMap<Double, double[][]>(colorToMaskMap);
		
		/* Translate the order to a simple list (for simplicity) */
		jaccardMaskListSorted_ = new ArrayList<double[][]>();
		for (double[][] mask : colorToMaskMapSorted.values()) {
			jaccardMaskListSorted_.add(mask);
		}
	}


	private static Map<Double, double[][]> computeColorToMaskMapping(String[] maskFilenames) {
		/* Build the list containing 2D image pixels */
		List<double[][]> binaryMaskList = new ArrayList<double[][]>();
		for (String filename : maskFilenames) {
			final String path = Paths.get(WORKING_DIR, MASK_FOLDER, filename).toString();
			final double[][] binaryMask = JaccardMaskManager.getMaskPixelsFromPath(path);
			binaryMaskList.add(binaryMask);
		}
		
		final int numMasks = binaryMaskList.size();
		
		Map<Double, double[][]> colorToMaskMap = new HashMap<Double, double[][]>();
		final double[][] markerPixels = ImageUtils.extractPixelsFromProcessor(
				watershedMarkers_, width_, height_, false);
		
		// loop through all marker (seed) image pixels
		for (int i = 0 ; i < width_ ; i++) {
			for (int j = 0 ; j < height_ ; j++) {
				
				final double colorValue = markerPixels[i][j];
				
				// pixels other than seeds are plain black
				final boolean pixelIsSeed = colorValue > 0;
				
				if (pixelIsSeed) {
					
					// find the mask that contains this seed
					for (int k = 0 ; k < numMasks ; k++) {
						
						final double[][] mask = binaryMaskList.get(k);
						final boolean pixelInMask = mask[i][j] > 0;
						
						if (pixelInMask) {
							// mask found, link the color value to that mask
							colorToMaskMap.put(colorValue, mask);
						}
						
					}
					
				}
				
			}
		}
		return colorToMaskMap;
	}


	private static void computeJaccardStatistics(String filename, 
			double runtimeSeconds, boolean outputCsv) throws IOException {
		
		/* Average */
		double jaccardAverage = 0;
		Map<String, Double> allJaccards = new HashMap<String, Double>();
		for (int i = 0 ; i < numCells_ ; i++) {
			
			final double[][] binaryCell = 
					ImageUtils.convert1DTo2DImage(resultCells_[i].getPixels(), width_, height_, true);
			
//			final double[][] watershedMarkers = 
//					ImageUtils.getPixelsFromPath(watershedMarkersPath_, false);
//			final double[][] binaryMask = 
//					WatershedUtils.findCorrespondingJaccardMask(
//							binaryMaskList_, binaryCell, watershedMarkers);
			final double[][] binaryMask = jaccardMaskListSorted_.get(i);

			final double jaccardIdx = JaccardMasks.computeJaccard(binaryCell, binaryMask);
			System.out.println(String.format("Jaccard index for cell %d : %.3f", 
					i+1, jaccardIdx));
			jaccardAverage += jaccardIdx;
			allJaccards.put(String.valueOf(i), jaccardIdx);
		}
		jaccardAverage /= (double) numCells_;
		
		/* Standard deviation */
		double jaccardVariance = 0;
		for (Double jIdx : allJaccards.values()) {
			jaccardVariance += Math.pow(jIdx - jaccardAverage, 2);
		}
		jaccardVariance /= (double) numCells_;
		final double jaccardStd = Math.sqrt(jaccardVariance);
		
		System.out.println(String.format("Averaged Jaccard index for image %s : %.3f", 
				filename, jaccardAverage));
		
		if (outputCsv) {
			testWriter_.outputJaccardStatistics(filename, jaccardAverage, jaccardStd,
					allJaccards, runtimeSeconds);
		}
	}

	
	private static String getDateTime() {
		DateFormat dateFormat = new SimpleDateFormat("yyMMdd-HHmm");
		return dateFormat.format(new Date());
	}
}
