package plugins.big.bigsnake.tests.watersheds;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import ij.process.ImageProcessor;
import plugins.big.bigsnake.utils.ImageUtils;

public class WatershedUtils {

	/**
	 * Supposed to find the region mask that gives the best Jaccard index 
	 * among all masks found in the directory.
	 */
	public static double[][] findCorrespondingJaccardMask(List<double[][]> maskList, 
			double[][] binaryRegion, double[][] seedImage) {
		
		final int width = seedImage.length;
		final int height = seedImage[0].length;
		
		// loop over pixels of the seed image
		for (int i = 0 ; i < width ; i++) {
			for (int j = 0 ; j < height ; j++) {
				
				final boolean pixelIsSeed = seedImage[i][j] > 0;
				
				if (pixelIsSeed) {
					
					/* once we found a seed, find which Jaccard mask corresponds
					 * to the binary region from watersheds that contains the seed 
					 */
					
					for (double[][] binaryMask : maskList) {
						
						final boolean pixelIsInMask = binaryMask[i][j] > 0;
						final boolean pixelIsInRegion = binaryRegion[i][j] > 0;
								
						if (pixelIsInMask && pixelIsInRegion) {
							
							return binaryMask;
							
						}
					}
				}
			}
		}
		System.err.println("No corresponding Jaccard mask for provided watershed region");
		return null;
	}

	public static ImageProcessor[] separateCells(ImageProcessor watershedResult, 
			double colorToExclude, boolean colorSort) {
		
		double[] pixels = ImageUtils.extractPixelsFromProcessor(watershedResult);
		final int size = pixels.length;
		
		final List<Double> colorValues = getRegionValues(pixels, colorToExclude, colorSort);
		final int numRegions = colorValues.size();
		
		List<float[]> regionImagePixelsList = new ArrayList<float[]>();
		for (int i = 0 ; i < numRegions ; i++) {
			float[] newImagePixels = new float[size];
			Arrays.fill(newImagePixels, 0); // fill it with black (binary 0)
			regionImagePixelsList.add(newImagePixels);
		}
		
		for (int i = 0 ; i < size ; i++) {
			final int colorIdx = colorValues.indexOf(pixels[i]);
			if (colorIdx > -1) {
				// if inside a region, set the pixel to white (binary 1)
				regionImagePixelsList.get(colorIdx)[i] = 1;
			}
		}
		
		// convert images into ImageProcessor objects (easier to save on disk)
		ImageProcessor[] resultCells = new ImageProcessor[numRegions];
		for (int i = 0 ; i < numRegions ; i++) {
			final ImageProcessor ip = watershedResult.createProcessor(
					watershedResult.getWidth(), watershedResult.getHeight());
			ip.setPixels(regionImagePixelsList.get(i));
			ip.resetMinAndMax();
			resultCells[i] = ip;
		}
		return resultCells;
	}

	public static List<Double> getRegionValues(double[] pixels, 
			double colorToExclude, boolean sort) {
		
		List<Double> regionColorValues = new ArrayList<Double>();
		
		for (int i = 0 ; i < pixels.length ; i++) {
			
			final double pix = pixels[i];
			boolean excluded = false;
			if (colorToExclude > 0) excluded = (pix == colorToExclude);
			
			if (!regionColorValues.contains(pix) && !excluded) {
				regionColorValues.add(pix);
			}
		}
		
		if (sort) {
			Collections.sort(regionColorValues);
		}
		
		return regionColorValues;
	}

}
