package plugins.big.bigsnake.voronoi;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;



/**
 * Some geometry utility methods (subclasses Double and Integer).
 * 
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 *
 * @version November 28, 2017
 */
public abstract class GeometryUtils {

	public static class Double extends GeometryUtils {
		
		/** Constructor. */
		public Double() {}
		
		
		/**
		 * https://en.wikipedia.org/wiki/Centroid#Centroid_of_a_polygon
		 */
		public static Point2D getPolygonCentroid(List<Point2D> points) {
			int N = points.size();
			List<Point2D> pointsL = new ArrayList<Point2D>(points);
			
			// make sure that it is a closed loop (idx 0 = idx N)
			if (points.get(N-1) != points.get(0)) {
				pointsL.add(points.get(0));
				N++;
			}
			
			double areaTimes6 = 6 * getPolygonArea(points);
			double cx = 0;
			double cy = 0;
			for (int i = 0 ; i < N-1 ; i++) {
				double factor = (pointsL.get(i).getX() * pointsL.get(i+1).getY()
						- pointsL.get(i+1).getX() * pointsL.get(i).getY());
				cx += (pointsL.get(i).getX() + pointsL.get(i+1).getX()) * factor;
				cy += (pointsL.get(i).getY() + pointsL.get(i+1).getY()) * factor;
			}
			cx /= areaTimes6;
			cy /= areaTimes6;
//			System.out.println("cx = " + cx + ", cy = " + cy + ", 6*area = " + areaTimes6);
			return new Point2D.Double(cx, cy);
		}
		
		
		/**
		 * https://en.wikipedia.org/wiki/Centroid#Centroid_of_a_polygon
		 */
		public static double getPolygonArea(List<Point2D> points) {
			int N = points.size();
			List<Point2D> pointsL = new ArrayList<Point2D>(points);
			
			// make sure that it is a closed loop (idx 0 = idx N)
			if (points.get(N-1) != points.get(0)) {
				pointsL.add(points.get(0));
				N++;
			}
			
			double area = 0;
			for (int i = 0 ; i < N-1 ; i++) {
				area += pointsL.get(i).getX() * pointsL.get(i+1).getY() 
						- pointsL.get(i+1).getX() * pointsL.get(i).getY();
			}
			area *= 0.5;
			return area;
		}
		
		
		public static boolean isClosed(List<Point2D.Double> points) {
			return (points.get(0) == points.get(points.size()-1));
		}
	}
	


	public static class Integer extends GeometryUtils {
		
		/** Constructor. */
		public Integer() {}
		
		
		/**
		 * https://en.wikipedia.org/wiki/Centroid#Centroid_of_a_polygon
		 */
		public static Point getPolygonCentroid(List<Point> points) {
			int N = points.size();
			List<Point> pointsL = new ArrayList<Point>(points);
			
			// make sure that it is a closed loop (idx 0 = idx N)
			if (points.get(N-1) != points.get(0)) {
				pointsL.add(points.get(0));
				N++;
			}
			
			int areaTimes6 = 6 * getPolygonArea(points);
			int cx = 0;
			int cy = 0;
			for (int i = 0 ; i < N-1 ; i++) {
				double factor = (pointsL.get(i).getX() * pointsL.get(i+1).getY()
						- pointsL.get(i+1).getX() * pointsL.get(i).getY());
				cx += (pointsL.get(i).getX() + pointsL.get(i+1).getX()) * factor;
				cy += (pointsL.get(i).getY() + pointsL.get(i+1).getY()) * factor;
			}
			cx /= areaTimes6;
			cy /= areaTimes6;
//			System.out.println("cx = " + cx + ", cy = " + cy + ", 6*area = " + areaTimes6);
			return new Point(cx, cy);
		}
		
		
		/**
		 * https://en.wikipedia.org/wiki/Centroid#Centroid_of_a_polygon
		 */
		public static int getPolygonArea(List<Point> points) {
			int N = points.size();
			List<Point> pointsL = new ArrayList<Point>(points);
			
			// make sure that it is a closed loop (idx 0 = idx N)
			if (points.get(N-1) != points.get(0)) {
				pointsL.add(points.get(0));
				N++;
			}
			
			int area = 0;
			for (int i = 0 ; i < N-1 ; i++) {
				area += pointsL.get(i).getX() * pointsL.get(i+1).getY() 
						- pointsL.get(i+1).getX() * pointsL.get(i).getY();
			}
			area *= 0.5;
			return area;
		}
		
		
		public static List<Point2D.Double> dilatePolygon(List<Point> points,
				Point centroid, double dilationFactorX, double dilationFactorY) {
			List<Point2D.Double> dilatedPoints = new ArrayList<Point2D.Double>();
			
			double fx = dilationFactorX;
			double fy = dilationFactorY;
			
			int cx = centroid.x;
			int cy = centroid.y;
			
			for (Point point : points) {
				double dilatedXCoord = fx * point.x + (1 - fx) * cx;
				double dilatedYCoord = fy * point.y + (1 - fy) * cy;
				dilatedPoints.add(new Point2D.Double(dilatedXCoord, dilatedYCoord));
			}
			
			return dilatedPoints;
		}


		public static boolean isClosed(List<Point> points) {
			return (points.get(0) == points.get(points.size()-1));
		}
	}
}

