package plugins.big.bigsnake.voronoi;


import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;

import flanagan.math.DeepCopy;
import kn.uni.voronoitreemap.datastructure.OpenList;
import kn.uni.voronoitreemap.diagram.PowerDiagram;
import kn.uni.voronoitreemap.j2d.PolygonSimple;
import kn.uni.voronoitreemap.j2d.Site;

/**
 * Class that handles the search of initial GlobalSnake control points,
 * through the computation of the Voronoi diagram, whose seeds are cell nuclei.
 * 
 * It uses the open-source efficient implementation found here 
 * Power Voronoi Diagram : <a>https://github.com/ArlindNocaj/power-voronoi-diagram</a>
 * 
 * @version January 13, 2018
 * 
 * @author Aymeric Galan (aymeric.galan@gmail.com)
 *
 */
public class VoronoiDiagram {
	
	private int numSites_ = 0;
	
	private int width_ = 0;
	private int height_ = 0; 
	
	private PolygonSimple boundingPolygon_ = null;
	
	private OpenList sites_ = null;
	
	private PolygonSimple[] diagramPolygons_ = null;

	private PowerDiagram diagram_ = null;
	
	private boolean computed_ = false;
	
	private double convexHullDilationCoef_ = 0;
	
	// debug purposes, make the bounding polygon to be the image boundaries
	private final boolean SIMPLE_BOUNDS = false;
	
	//--------------------
	// PUBLIC METHODS
	
	/** Constructor. */
	public VoronoiDiagram(int width, int height,
			Point2D.Double[] sitesPositions, double dilationCoef) {
		width_  = width;
		height_ = height;
		numSites_ = sitesPositions.length;
		convexHullDilationCoef_ = dilationCoef;
		
		sites_ = new OpenList();
		for (int i = 0 ; i < numSites_ ; i++) {
			Site site = new Site(sitesPositions[i].getX(), 
					sitesPositions[i].getY());
			sites_.add(site);
		}
		
		// build polygon that will limit the extent of the Voronoi diagram
		buildBoundingPolygon();
		
		/* initialize the diagram */
		diagram_ = new PowerDiagram(sites_, boundingPolygon_);
		
		computed_ = false;
	}
	
	
	public void runComputation(boolean display) {
		
	    // perform the computation
	    diagram_.computeDiagram();
	    
	    // for each site we can now get the resulting polygon of its cell. 
	    // note that the cell can also be empty, in this case there is no polygon for the corresponding site.
	    diagramPolygons_ = new PolygonSimple[numSites_];
	    for (int i = 0 ; i < numSites_ ; i++) {
	    	Site site = sites_.array[i];

	    	PolygonSimple polygon = new PolygonSimple();
	    	for (kn.uni.voronoitreemap.j2d.Point2D pt : site.getPolygon()) {
	    		polygon.add(pt.x, pt.y);
	    	}
	    	
	    	diagramPolygons_[i] = polygon;
	    }
	    
	    // show a (very buggy inside Icy, better from Eclipse) representation of the diagram
	    if (display) {
	      diagram_.showDiagram();
	    }
	    computed_ = true;
	}
	
	
	public PolygonSimple[] getPolygons() {
		return diagramPolygons_;
	}
	
	
	public Point2D.Double[] getSitesAsPoints() {
		Point2D.Double[] points = new Point2D.Double[numSites_];
		for (int i = 0 ; i < numSites_ ; i++) {
			Site site = sites_.get(i);
			points[i] = new Point2D.Double(site.x, site.y);
		}
		return points;
	}
	
	public boolean isComputed() {
		return computed_;
	}
	
	//------------------
	// PRIVATE METHODS

	/**
	 * Compute the bounding polygon of the diagram
	 * as the convex hull of the set composed of all the sites,
	 * then scale it up.
	 */
	private void buildBoundingPolygon() {
		if (SIMPLE_BOUNDS) {
			boundingPolygon_ = new PolygonSimple(4);
			boundingPolygon_.add(0, 0);
			boundingPolygon_.add(width_, 0);
			boundingPolygon_.add(width_, height_);
			boundingPolygon_.add(0, height_);
		} 
		/* Dilate the convex hull to build the bounding polygon. */
		else if (convexHullDilationCoef_ != 0) {
			// compute the convex hull
			List<Point> convexHull = 
					GrahamScan.getConvexHull(getSitesX(), getSitesY());
			
			// then dilate the polygon
			final double dilationCoefX = convexHullDilationCoef_;
			final double dilationCoefY = convexHullDilationCoef_;
			Point centroid = GeometryUtils.Integer.getPolygonCentroid(convexHull);
			List<Point2D.Double> convexHullDilated = 
					GeometryUtils.Integer.dilatePolygon(convexHull, 
							centroid, dilationCoefX, dilationCoefY);
			
			int nPoints = GeometryUtils.Double.isClosed(convexHullDilated) ? 
					convexHullDilated.size() - 1 : convexHullDilated.size();
					
			boundingPolygon_ = new PolygonSimple(nPoints);
			for (int i = 0 ; i < nPoints ; i++) {
				Point2D.Double point = convexHullDilated.get(i);
				boundingPolygon_.add(point.getX(), point.getY());
			}
		} else {
			System.err.println("VoronoiDiagram : Bounding polygon not implemented !");
			return;
		}
	}
	
	
	private int[] getSitesX() {
		int[] allX = new int[numSites_];
		for (int i = 0 ; i < numSites_ ; i++) {
			allX[i] = (int) sites_.array[i].x;
		}
		return allX;
	}
	
	
	private int[] getSitesY() {
		int[] allY = new int[numSites_];
		for (int i = 0 ; i < numSites_ ; i++) {
			allY[i] = (int) sites_.array[i].y;
		}
		return allY;
	}

	
//	/** Utility function to convert OpenList (of vertices) to List<Point2D.Double>. */
//	private static List<Point2D.Double> openList2ListPoint(OpenList olist) {
//		int N = olist.size;
//		double[] allX = new double[N];
//		double[] allY = new double[N];
//		
//		List<Point2D.Double> list = new ArrayList<Point2D.Double>();
//		for (int i = 0 ; i < N ; i++) {
//			Site site = olist.get(i);
//			list.add(new Point2D.Double(site.x, site.y));
//		}
//		return list;
//	}
	
}